import time
from typing import Callable, Dict, List, Tuple

import GPy
import nearestPD
import numpy as np
import scipy.linalg as la
from GPy.inference.latent_function_inference.expectation_propagation import posteriorParams  # , gaussianApproximation
from GPy.inference.latent_function_inference.posterior import PosteriorEP as Posterior
from GPy.util.linalg import dpotrs, dtrtrs, jitchol, pdinv, symmetrify, tdot
from scipy.integrate import dblquad, quad
from scipy.linalg import inv, sqrtm, svd
from scipy.special import logsumexp
from scipy.stats import multivariate_normal, norm

from .. import util

# Some helper functions:
npdf = lambda x, m, v: 1.0 / np.sqrt(2 * np.pi * v) * np.exp(-((x - m) ** 2) / (2 * v))
log_npdf = lambda x, m, v: -0.5 * np.log(2 * np.pi * v) - (x - m) ** 2 / (2 * v)
phi = lambda x: norm.cdf(x)
logphi = lambda x: norm.logcdf(x)

log_2_pi = np.log(2 * np.pi)


def posdef_sqrtm(M: np.ndarray) -> np.ndarray:
    """
    Returns a square root of a positive definite matrix
    :param M: A positive definite matrix
    :return: Squarte root of M
    """
    (U, S, VT) = svd(M)
    D = np.diag(np.sqrt(S))
    return U @ D @ VT


def sqrtm_block(M: np.ndarray, y: List[Tuple[int, float]], yc: List[List[Tuple[int, int]]]) -> np.ndarray:
    """
    Returns a square root of a positive definite matrix
    :param M: A positive definite block matrix
    :param y: Observations indicating where we have a diagonal element
    :param yc: Comparisons indicating where we have a block diagonal element
    :return: Squarte root of M
    """
    Msqrtm = np.zeros(M.shape)
    if len(y) > 0:
        for yi, yval in y:
            Msqrtm[yi, yi] = np.sqrt(M[yi, yi])
    if len(yc) > 0:
        for ycb in yc:
            loc_inds_winners, loc_inds_loosers = [ycb[k][0] for k in range(len(ycb))], [
                ycb[k][1] for k in range(len(ycb))
            ]
            batch = np.sort(np.unique(loc_inds_winners + loc_inds_loosers))
            Msqrtm[np.ix_(batch, batch)] = posdef_sqrtm(M[np.ix_(batch, batch)])
    return Msqrtm


class MarginalMoments(object):
    """
    A simple container for the marginal moments
    :param num_data: how many observations are there
    """

    def __init__(self, num_data):
        self.logZ_hat = np.empty(num_data, dtype=np.float64)
        self.mu_hat = np.empty(num_data, dtype=np.float64)
        self.sigma2_hat = np.empty((num_data, num_data), dtype=np.float64)


class GaussianApproximation(object):
    """
    A simple container for Gaussian approximations of the batches
    :param v: scale
    :param tau: precision
    """

    def __init__(self, v, tau):
        self.v = v
        self.tau = tau

    def _update_batch(
        self,
        eta: float,
        delta: float,
        post_params: posteriorParams,
        marg_moments: MarginalMoments,
        batch: List[int],
        get_logger: Callable = None,
        sigma2s: np.ndarray = None,
    ):
        """
        Computes new gaussian approximation for a batch given posterior and marginal moments. See e.g. 3.59 in http://www.gaussianprocess.org/gpml/chapters/RW.pdf
        :param eta: parameter for fractional updates.
        :param delta: damping updates factor.
        :param post_params: Posterior approximation
        :param marg_moments: Marginal moments at this iteration
        :param batch: list of indices of the parameters to be updated
        :param get_logger: Function for receiving the legger where the prints are forwarded.
        """
        sigma_hat_inv, _, _, _ = pdinv(marg_moments.sigma2_hat[np.ix_(batch, batch)])
        post_sigma_inv, _, _, _ = pdinv(post_params.Sigma[np.ix_(batch, batch)])

        tmp0 = sigma_hat_inv - post_sigma_inv

        delta_tau = delta / eta * tmp0
        delta_v = (
            delta
            / eta
            * (np.dot(marg_moments.mu_hat[batch], sigma_hat_inv) - np.dot(post_params.mu[batch], post_sigma_inv))
        )
        tau_tilde_prev = self.tau[np.ix_(batch, batch)]
        tmp = (1 - delta) * self.tau[np.ix_(batch, batch)] + delta_tau

        # Let us umake sure that sigma_hat_inv-post_sigma_inv is positive definite
        tmp, added_value = nearestPD.nearest_pd.nearestPD(tmp)
        update = True
        if (added_value > 1) and (sigma2s is not None):
            update = False
            sigma2s *= 1.05
            if get_logger is not None:
                get_logger().error(
                    "Increasing batch noise. Not updating gaussian approximation ({})".format(sigma2s[0])
                )
        if update:
            self.tau[np.ix_(batch, batch)] = tmp
            self.v[batch] = (1 - delta) * self.v[batch] + delta_v
        return (delta_tau, delta_v), sigma2s


class CavityParams(object):
    """
    A simple container for the cavity parameters
    :param num_data: how many observations are there
    """

    def __init__(self, num_data: int):
        self.v = np.zeros(num_data, dtype=np.float64)
        self.tau = np.zeros((num_data, num_data), dtype=np.float64)

    def _update_i(self, eta: float, ga_approx: GaussianApproximation, post_params: posteriorParams, i: int):
        """
        Computes the cavity params for specific index
        :param eta: parameter for fractional updates.
        :param ga_approx: Gaussian approximation
        :param post_params: Posterior approximation
        :param i: index of the parameters to be updated
        """
        self.tau[i, i] = 1.0 / post_params.Sigma[i, i] - eta * ga_approx.tau[i, i]
        self.v[i] = post_params.mu[i] / post_params.Sigma[i, i] - eta * ga_approx.v[i]

    def _update_batch(
        self,
        eta: float,
        ga_approx: GaussianApproximation,
        post_params: posteriorParams,
        batch: List[int],
        get_logger: Callable = None,
    ):
        """
        Computes the cavity params for specific batch
        :param eta: parameter for fractional updates.
        :param ga_approx: Gaussian approximation
        :param post_params: Posterior approximation
        :param batch: list of indices of the parameters to be updated
        :param get_logger: Function for receiving the legger where the prints are forwarded.
        """
        post_sigma_inv = inv(post_params.Sigma[np.ix_(batch, batch)])
        tmp = post_sigma_inv - eta * ga_approx.tau[np.ix_(batch, batch)]
        self.tau[np.ix_(batch, batch)], _ = nearestPD.nearest_pd.nearestPD(tmp)
        self.v[batch] = np.dot(post_sigma_inv, post_params.mu[batch]) - eta * ga_approx.v[batch]


def update_posterior(
    K: np.ndarray,
    v: np.ndarray,
    tau: np.ndarray,
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
    jitter: float = 1e-9,
    get_logger: Callable = None,
) -> posteriorParams:
    """
    Update the posterior approximation. See e.g. 3.59 in http://www.gaussianprocess.org/gpml/chapters/RW.pdf
    :param K: prior covariance matrix
    :param v: Scale of the Gaussian approximation
    :param tau: Precision of the Gaussian approximation
    :param y: Observations indicating where we have a diagonal element
    :param yc: Comparisons indicating where we have a block diagonal element
    :param jitter: small number added to the diagonal to increase robustness.
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    :return: posterior approximation
    """
    D = K.shape[0]
    sqrt_tau = sqrtm_block(tau + np.diag(jitter * np.ones((D))), y, yc)
    G = np.dot(sqrt_tau, K)
    B = np.identity(D) + np.dot(G, sqrt_tau)
    L = jitchol(B)
    V = np.linalg.solve(L, G)
    Sigma_full = K - np.dot(V.T, V)
    mu = np.dot(Sigma_full, v)
    Sigma = np.diag(Sigma_full)

    return posteriorParams(mu=mu, Sigma=Sigma_full, L=L)


def ep_comparison(
    X: np.ndarray,
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
    kern: GPy.kern.Kern,
    sigma2s: np.ndarray,
    max_itt: int = 50,
    delta: float = 0.9,
    eta: float = 1.0,
    tol: float = 1e-6,
    ga_approx_old: GaussianApproximation = None,
    get_logger: Callable = None,
) -> Tuple[Posterior, int, Dict, GaussianApproximation]:
    """
    :param X: All locations of both direct observations and batch comparisons
    :param y: Direct observations as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param kern: GP kernel
    :param sigma2s: noise variance of observations
    :param max_itt: maximum number of iterations
    :param delta: damping updates factor.
    :param eta: parameter for fractional updates.
    :param tol: tolerance after which the EP is stopped unless too many iterations have passed
    :param ga_approx_old: If there has been previous gaussian approximation, it should be passed
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    :return: A tuple consisting of the posterior approximation, log marginal likelihood, radient dictionary and gaussian approximations of the batches
    """

    t0 = time.time()

    N = X.shape[0]
    Ndir = len(y)
    Ncomp = len(yc)
    ###################################################################################
    # Contruct observations and kernels
    ###################################################################################
    K = kern.K(X)

    ###################################################################################
    # Prepare marginal moments, site approximation and cavity containers
    ###################################################################################

    f_marg_moments = MarginalMoments(N)
    f_ga_approx = GaussianApproximation(np.zeros(N, dtype=np.float64), np.zeros((N, N), dtype=np.float64))
    f_cavity = CavityParams(N)

    # insert likelihood information to each gaussian approximation
    for i in range(Ndir):
        (ii, yi) = y[i]  # index in kernel, y value
        f_ga_approx.v[ii] = yi / sigma2s[i]
        f_ga_approx.tau[ii, ii] = 1.0 / sigma2s[i]

    if ga_approx_old is not None:  # If there exists old gaussian approximation, we reuse it
        N_old = ga_approx_old.tau.shape[0]
        if N - N_old > -1:
            f_ga_approx.v[:N_old] = ga_approx_old.v
            f_ga_approx.tau[np.ix_(np.arange(N_old), np.arange(N_old))] = ga_approx_old.tau

    ###################################################################################
    # Prepare global approximations
    ###################################################################################
    f_post_params = update_posterior(K, f_ga_approx.v, f_ga_approx.tau, y, yc)
    if np.any(np.isnan(f_post_params.mu)):
        if get_logger is not None:
            get_logger().error("Posterior mean contains nan in the EP approximation")

    ###################################################################################
    # Iterate
    ###################################################################################
    for itt in range(max_itt):
        old_params = np.hstack((f_post_params.mu.copy(), f_post_params.Sigma_diag.copy()))

        if get_logger is not None:
            get_logger().info("Iteration %d" % (itt + 1))
        d_list = []
        if len(yc) > 0:
            d_list = np.random.choice(range(len(yc)), size=len(yc), replace=False)
        for d in d_list:  # iterate through batches
            loc_inds_winners, loc_inds_loosers = [yc[d][k][0] for k in range(len(yc[d]))], [
                yc[d][k][1] for k in range(len(yc[d]))
            ]
            loc_inds_batch = np.sort(np.unique(loc_inds_winners + loc_inds_loosers))
            # get relevant EP parameters for comparison points
            f_cavity._update_batch(
                eta=eta, ga_approx=f_ga_approx, post_params=f_post_params, batch=loc_inds_batch, get_logger=get_logger
            )

            try:
                # get cavity parameters of the batch
                ind_winners, ind_loosers = [np.where(loc_inds_batch == it)[0][0] for it in loc_inds_winners], [
                    np.where(loc_inds_batch == it)[0][0] for it in loc_inds_loosers
                ]  # indices within a batch
                (
                    f_marg_moments.logZ_hat[loc_inds_batch],
                    f_marg_moments.mu_hat[loc_inds_batch],
                    f_marg_moments.sigma2_hat[np.ix_(loc_inds_batch, loc_inds_batch)],
                    sigma2s[loc_inds_batch],
                ) = _match_moments_batch(
                    f_cavity.v[loc_inds_batch],
                    f_cavity.tau[np.ix_(loc_inds_batch, loc_inds_batch)],
                    ind_winners,
                    ind_loosers,
                    sigma2s[loc_inds_batch],
                    N=100000,
                    get_logger=get_logger,
                )
            except AssertionError as e:
                if get_logger is not None:
                    get_logger().error("Numerical problem with feedback %d in iteration %d. Skipping update" % (d, itt))
            _, sigma2s[loc_inds_batch] = f_ga_approx._update_batch(
                eta=eta,
                delta=delta,
                post_params=f_post_params,
                marg_moments=f_marg_moments,
                batch=loc_inds_batch,
                get_logger=get_logger,
                sigma2s=sigma2s[loc_inds_batch],
            )
        f_post_params = update_posterior(K, f_ga_approx.v, f_ga_approx.tau, y, yc, get_logger=get_logger)

        if np.any(np.isnan(f_post_params.mu)) or np.any(np.isnan(f_post_params.mu)):
            if get_logger is not None:
                get_logger().error("Posterior mean contains nan in the EP approximation")

        # check for convergence
        new_params = np.hstack((f_post_params.mu.copy(), f_post_params.Sigma_diag.copy()))
        converged = True
        if np.mean((new_params - old_params) ** 2) / np.mean(old_params**2) < tol:
            pass
        else:
            converged = False
        if converged:
            run_time = time.time() - t0
            if get_logger is not None:
                get_logger().info("Converged in %d iterations in %4.3fs" % (itt + 1, run_time))
            break
    #############################################################################3
    # Marginal likelihood & gradients
    #############################################################################3

    # compute normalization constant for likelihoods
    for i in range(Ndir):
        (ii, yi) = y[i]  # index in kernel, y value
        f_cavity._update_i(eta=eta, ga_approx=f_ga_approx, post_params=f_post_params, i=ii)
        f_marg_moments.logZ_hat[ii] = log_npdf(
            yi, f_cavity.v[ii] / f_cavity.tau[ii, ii], 1.0 / f_cavity.tau[ii, ii] + sigma2s[i]
        )

    # marginal likelihood and gradient contribution from each f

    Z_tilde = _log_Z_tilde(f_marg_moments, f_ga_approx, f_cavity, y, yc)
    f_posterior, f_logZ, f_grad = _inference(K, f_ga_approx, f_cavity, Z_tilde, y, yc)
    return f_posterior, f_logZ, f_grad, f_ga_approx


def _match_moments_batch(
    v_cav: np.ndarray,
    tau_cav: np.ndarray,
    ind_winners: List[int],
    ind_loosers: List[int],
    nu2: np.ndarray,
    N: int = 10000,
    get_logger: Callable = None,
) -> Tuple[int, np.ndarray, np.ndarray]:
    """
    Computes the moments of the given batch
    :param v_cav: Scale of the cavity distribution
    :param tau_cav: Precision of the cavity distribution
    :param ind_winners: Indices of winners in each pairwise comparison of the batch
    :param ind_loosers: Indices of loosers in each pairwise comparison of the batch
    :param nu2: Noise variance of each batch member likelihood
    :param N: Number of parameters used to compute the moments
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    :return: Tuple containing the integral, mean and variance
    """
    n = tau_cav.shape[0]
    tau_cav_inv = inv(tau_cav)
    m_cav, sigma_cav = np.dot(v_cav, tau_cav_inv), tau_cav_inv
    logZ, site_m, site_m2 = _compute_moments_sampling(
        m_cav, sigma_cav, ind_winners, ind_loosers, nu2, N=N, get_logger=get_logger
    )

    min_eig0 = np.min(np.real(la.eig(site_m2)[0]))
    if min_eig0 < 0:
        site_m2 -= min_eig0 * np.eye(site_m2.shape[0])
        if get_logger is not None:
            get_logger().error(
                "marginal moment not suitable in moment matching, adding smallest eigenvalue ({}) to the diagonal to guarantee PD".format(
                    -1 * min_eig0
                )
            )

    if np.isnan(logZ) or np.any(np.isnan(site_m)) or np.any(np.isnan(site_m2)):
        raise AssertionError()
    return (
        np.array([logZ / n for i in range(n)]),
        site_m.reshape(
            n,
        ),
        site_m2,
        nu2,
    )


def _log_Z_tilde(
    marg_moments: MarginalMoments,
    ga_approx: GaussianApproximation,
    cav_params: CavityParams,
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
) -> float:
    """
    Give the log marginal likelihood of the posterior approximation
    :param marg_moments: Marginal moments of the posterior
    :param ga_approx: Gaussian approximation of the batches
    :param cav_params: Cavity parameters of the posterior
    :param y: Direct observations as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :return: log marginal likleihood of the posterior
    """

    # go through direct observations:
    inds = [np.array([y[i][0]]) for i in range(len(y))]  # direct observations
    for d in range(len(yc)):  # iterate through batches
        loc_inds_winners, loc_inds_loosers = [yc[d][k][0] for k in range(len(yc[d]))], [
            yc[d][k][1] for k in range(len(yc[d]))
        ]
        inds += [np.sort(np.unique(loc_inds_winners + loc_inds_loosers))]
    log_Z_tilde = np.sum(marg_moments.logZ_hat + 0.5 * np.log(2 * np.pi))
    for ind in inds:
        log_Z_tilde += (
            0.5
            * np.log(
                np.linalg.det(
                    np.eye(len(ind)) + np.dot(ga_approx.tau[np.ix_(ind, ind)], inv(cav_params.tau[np.ix_(ind, ind)]))
                )
            )
            - 0.5
            * np.linalg.multi_dot(
                [
                    ga_approx.v[ind].T,
                    inv(cav_params.tau[np.ix_(ind, ind)] + ga_approx.tau[np.ix_(ind, ind)]),
                    ga_approx.v[ind],
                ]
            )
            + 0.5
            * np.linalg.multi_dot(
                [
                    cav_params.v[ind].T,
                    inv(cav_params.tau[np.ix_(ind, ind)] + ga_approx.tau[np.ix_(ind, ind)]),
                    np.linalg.multi_dot(
                        [ga_approx.tau[np.ix_(ind, ind)], inv(cav_params.tau[np.ix_(ind, ind)]), cav_params.v[ind]]
                    )
                    - 2.0 * ga_approx.v[ind],
                ]
            )
        )
    return log_Z_tilde


def _ep_marginal(
    K: np.ndarray,
    ga_approx: GaussianApproximation,
    Z_tilde: float,
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
) -> Tuple[float, posteriorParams]:
    """
    Compute Gaussian log marginal and posterior
    :param K: prior covariance matrix
    :param ga_approx: Gaussian approximation of the batches
    :param Z_tilde: log marginal of the posterior
    :param y: Direct observations as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    """
    post_params = update_posterior(K, ga_approx.v, ga_approx.tau, y, yc)

    # Gaussian log marginal excluding terms that can go to infinity due to arbitrarily small tau_tilde.
    # These terms cancel out with the terms excluded from Z_tilde
    B_logdet = np.sum(2.0 * np.log(np.diag(post_params.L)))
    log_marginal = 0.5 * (
        -len(ga_approx.v) * log_2_pi - B_logdet + np.sum(ga_approx.v * np.dot(post_params.Sigma, ga_approx.v))
    )
    log_marginal += Z_tilde

    return log_marginal, post_params


def _inference(
    K: np.ndarray,
    ga_approx: GaussianApproximation,
    cav_params: CavityParams,
    Z_tilde: float,
    y: List[Tuple[int, float]],
    yc: List[List[Tuple[int, int]]],
) -> Tuple[Posterior, int, Dict]:
    """
    Compute the posterior approximation
    :param K: prior covariance matrix
    :param ga_approx: Gaussian approximation of the batches
    :param cav_params: Cavity parameters of the posterior
    :param Z_tilde: Log marginal likelihood
    :param y: Direct observations as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :return: A tuple consisting of the posterior approximation, log marginal likelihood and gradient dictionary
    """

    log_marginal, post_params = _ep_marginal(K, ga_approx, Z_tilde, y, yc)
    tau_tilde_root = sqrtm_block(ga_approx.tau, y, yc)
    Sroot_tilde_K = np.dot(tau_tilde_root, K)
    aux_alpha, _ = dpotrs(post_params.L, np.dot(Sroot_tilde_K, ga_approx.v), lower=1)
    alpha = (ga_approx.v - np.dot(tau_tilde_root, aux_alpha))[:, None]  # (K + Sigma^(\tilde))^(-1) /mu^(/tilde)
    LWi, _ = dtrtrs(post_params.L, tau_tilde_root, lower=1)

    Wi = np.dot(LWi.T, LWi)
    symmetrify(Wi)  # (K + Sigma^(\tilde))^(-1)
    dL_dK = 0.5 * (tdot(alpha) - Wi)
    dL_dthetaL = 0
    return (
        Posterior(woodbury_inv=np.asfortranarray(Wi), woodbury_vector=alpha, K=K),
        log_marginal,
        {"dL_dK": dL_dK, "dL_dthetaL": dL_dthetaL, "dL_dm": alpha},
    )


def _compute_moments_sampling(
    mf: np.ndarray,
    vf: np.ndarray,
    ind_winners: List[int],
    ind_loosers: List[int],
    nu2: np.ndarray,
    N: int = 10000,
    get_logger: Callable = None,
) -> Tuple[float, np.ndarray, np.ndarray]:
    """
    Compute the marginal moments of a batch using numerical integration
    :param mf: cavity mean of the batch
    :params vf: cavity covariance of the batch
    :params ind_winners: indices of comparison winners
    :params ind_loosers: indices of comparison loosers
    :param nu2: noise variance of the batch members
    :param N: number of samples used for approximating the moments
    :param get_logger: Function for receiving the logger where the prints are forwarded.
    :return: Tuple containing the integral, first and second moments
    """
    m = vf.shape[0]
    sqrt_nu = np.sqrt(nu2)

    f = np.random.multivariate_normal(mf, vf, N)

    # Generate samples to compute likelihood
    if not (np.prod(ind_winners == ind_winners[0]) == 1):  # batch winner case
        return

    ind_winner = ind_winners[0]
    s = f[:, ind_winner] + sqrt_nu[ind_winner] * np.random.randn(N)
    s_compare = phi((f[:, ind_loosers] - s[:, None]) / sqrt_nu[None, ind_loosers, 0])
    s_compare = np.exp(np.log(s_compare).sum(axis=1))
    likelihood = s_compare.copy()
    qs0_ = np.mean(likelihood)
    qs0_ = max([qs0_, 1e-9])

    # First moments:
    f_nonzero = likelihood[:, None] * f
    likelihood_sqrt = np.sqrt(likelihood)
    f_nonzero_sqrt = likelihood_sqrt[:, None] * f
    qs1_ = np.sum(f_nonzero, axis=0) / qs0_ / N

    # Second moments
    diff_f = f_nonzero_sqrt - qs1_[None, :] * likelihood_sqrt[:, None]
    qs2_ = diff_f.T @ diff_f / qs0_ / N

    return qs0_, qs1_, qs2_



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import logging
from typing import Callable, List, Union

import numpy as np

from ..event_handler import EventHandler
from .candidate_point_calculators import CandidatePointCalculator
from .loop_state import LoopState
from .model_updaters import ModelUpdater
from .stopping_conditions import ConvergenceStoppingCondition, FixedIterationsStoppingCondition, StoppingCondition
from .user_function import UserFunction, UserFunctionWrapper
from .user_function_result import UserFunctionResult

_log = logging.getLogger(__name__)


class OuterLoop(object):
    """
    Generic outer loop that provides the framework for decision making parts of Emukit.

    The loop can be used in two modes:

    1. Emukit calculates the next point(s) to try and evaluates your function at these points until some stopping
       criterion is met.
    2. Emukit only calculates the next points(s) to try and you evaluate your function or perform the experiment.

    This object exposes the following events. See ``emukit.core.event_handler`` for details of how to subscribe:
         - ``loop_start_event`` called at the start of the `run_loop` method
         - ``iteration_end_event`` called at the end of each iteration
    """

    def __init__(
        self,
        candidate_point_calculator: CandidatePointCalculator,
        model_updaters: Union[ModelUpdater, List[ModelUpdater]],
        loop_state: LoopState = None,
    ) -> None:
        """
        :param candidate_point_calculator: Finds next points to evaluate by optimizing the acquisition function
        :param model_updaters: Updates the data in the model(s) and the model hyper-parameters when we observe new data
        :param loop_state: Object that keeps track of the history of the loop.
                           Default: None, resulting in empty initial state
        """
        self.candidate_point_calculator = candidate_point_calculator

        if isinstance(model_updaters, list):
            self.model_updaters = model_updaters
        else:
            self.model_updaters = [model_updaters]
        self.loop_state = loop_state
        if self.loop_state is None:
            self.loop_state = LoopState([])

        self.loop_start_event = EventHandler()
        self.iteration_end_event = EventHandler()

    def run_loop(
        self,
        user_function: Union[UserFunction, Callable],
        stopping_condition: Union[StoppingCondition, int],
        context: dict = None,
    ) -> None:
        """
        :param user_function: The function that we are emulating
        :param stopping_condition: If integer - a number of iterations to run, or an object - a stopping
                        condition object that decides whether we should stop collecting more points.
                        Note that stopping conditions can be logically combined (&, |)
                        to represent complex stopping criteria.
        :param context: The context is used to force certain parameters of the inputs to the function of interest to
                        have a given value. It is a dictionary whose keys are the parameter names to fix and the values
                        are the values to fix the parameters to.
        """

        is_int = isinstance(stopping_condition, int)
        is_single_condition = isinstance(stopping_condition, StoppingCondition)

        if not (is_int or is_single_condition):
            raise ValueError(
                "Expected stopping_condition to be an int or a StoppingCondition instance,"
                "but received {}".format(type(stopping_condition))
            )

        if not isinstance(user_function, UserFunction):
            user_function = UserFunctionWrapper(user_function)

        if isinstance(stopping_condition, int):
            stopping_condition = FixedIterationsStoppingCondition(stopping_condition + self.loop_state.iteration)

        _log.info("Starting outer loop")

        self.loop_start_event(self, self.loop_state)

        while not stopping_condition.should_stop(self.loop_state):
            _log.info("Iteration {}".format(self.loop_state.iteration))

            self._update_models()
            new_x = self.candidate_point_calculator.compute_next_points(self.loop_state, context)
            _log.debug("Next suggested point(s): {}".format(new_x))
            results = user_function.evaluate(new_x)
            _log.debug("User function returned: {}".format(results))
            self.loop_state.update(results)
            self.iteration_end_event(self, self.loop_state)

        self._update_models()
        _log.info("Finished outer loop")

    def _update_models(self):
        for model_updater in self.model_updaters:
            model_updater.update(self.loop_state)

    def get_next_points(self, results: List[UserFunctionResult] = None, context: dict = {}) -> np.ndarray:
        """
        This method is used when the user doesn't want Emukit to evaluate the function of interest but rather just wants
        the input locations to evaluate the function at. This method calculates the new input locations.

        :param results: Function results since last loop step
        :param context: A dictionary of fixed parameters, identical to the context used in
                        self.run_loop()
        :return: Next batch of points to run
        """
        if results:
            self.loop_state.update(results)
            self._update_models()
        new_x = self.candidate_point_calculator.compute_next_points(self.loop_state, context)
        return new_x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .bandit_parameter import BanditParameter  # noqa: F401
from .categorical_parameter import CategoricalParameter  # noqa: F401
from .continuous_parameter import ContinuousParameter  # noqa: F401
from .discrete_parameter import DiscreteParameter, InformationSourceParameter  # noqa: F401
from .encodings import OneHotEncoding, OrdinalEncoding  # noqa: F401
from .parameter import Parameter  # noqa: F401
from .parameter_space import ParameterSpace  # noqa: F401



from typing import Callable, Dict, Optional, Tuple

import numpy as np

import emukit
from emukit.core import ParameterSpace
from emukit.core.acquisition import Acquisition
from emukit.core.interfaces import IModel
from emukit.core.optimization import ContextManager, GradientAcquisitionOptimizer

from .. import ComparisonGP, ComparisonGPEmukitWrapper


class AcquisitionFunction:
    """
    This class implements the general AcquisitionFunction

    :param options: Dictionary containing the acquisition function options
    :param optimizer_options: Dictionary containing options for the acquisition function optimizer
    """

    def __init__(self, options: Dict = {}, optimizer_options: Dict = {}):
        self.optimizer_options = optimizer_options
        self.pool_size = options.get("pool_size", -1)
        self.acq_samples = options.get("acq_samples", 1000)
        self.acq_opt_restarts = options.get("acq_opt_restarts", 10)

    def acq_fun_optimizer(
        self, m: ComparisonGP, bounds: np.ndarray, batch_size: int, get_logger: Callable
    ) -> np.ndarray:
        """
        Implements the optimization scheme for the acquisition function

        :param m: The model which posterior is used by the acquisition function (from which the samples are drawn from)
        :param bounds: the optimization bounds of the new sample
        :param batch_size: How many points are there in the batch
        :param get_logger: Function for receiving the legger where the prints are forwarded.
        :return: optimized locations
        """
        raise NotImplementedError

    def reset(self, model: ComparisonGPEmukitWrapper) -> None:
        """
        Some acquisition functions need to be reseted, this method is for that.
        :param model: the model to be passed to the acquisition function (some acquisition functions need a model at this point)
        """
        None


class EmukitAcquisitionFunctionWrapper(Acquisition):
    def __init__(self, model: ComparisonGP, acquisitionFunction: AcquisitionFunction) -> None:
        """
        This acquisition computes for a given input the improvement over the current best observed value in
        expectation. For more information see:
        Efficient Global Optimization of Expensive Black-Box Functions
        Jones, Donald R. and Schonlau, Matthias and Welch, William J.
        Journal of Global Optimization
        :param model: model that is used to compute the improvement.
        :param acquisitionFunction: The used acquisitionFunction
        """

        self.model = model
        self.acquisitionFunction = acquisitionFunction
        self.size = self.model.X.shape[0]

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the Expected Improvement.
        :param x: points where the acquisition is evaluated.
        :return: acquisition function values
        """
        return -self.acquisitionFunction.evaluate(x, self.model)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the Expected Improvement.
        :param x: points where the acquisition is evaluated.
        :return: a tuple containing the acquisition function values and their gradients
        """
        f, g = self.acquisitionFunction.evaluate_with_gradients(x, self.model)
        return -f, -g

    def reset(self, model: ComparisonGPEmukitWrapper) -> None:
        """
        Some acquisition functions need to be reseted, this method is for that.
        :param model: the model to be passed to the acquisition function (some acquisition functions need a model at this point)
        """
        self.acquisitionFunction.reset(model.model)

    @property
    def has_gradients(self) -> bool:
        """
        :return: True as all acquisition functions have gradients
        """
        return True


class SequentialGradientAcquisitionOptimizer(GradientAcquisitionOptimizer):
    def __init__(self, space: ParameterSpace, batch_size: int) -> None:
        """
        :param space: The parameter space spanning the search problem.
        """
        self.batch_size = batch_size
        super().__init__(space)

    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager) -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of abstract method.
        Taking into account gradients if acquisition supports them.
        See AcquisitionOptimizerBase._optimizer for parameter descriptions.
        See class docstring for implementation details.
        """
        x_min, fx_min = None, None
        for i in range(self.batch_size):
            acquisition.reset(acquisition.model)
            x_min_i, fx_min_i = super()._optimize(acquisition, context_manager)

            if x_min is None:
                x_min, fx_min = x_min_i, fx_min_i
            else:
                x_min, fx_min = np.concatenate((x_min, x_min_i), axis=1), fx_min + fx_min_i
        return x_min, fx_min

    def optimize(self, acquisition: Acquisition, context: Optional = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Optimizes the acquisition function.
        :param acquisition: The acquisition function to be optimized
        :param context: Optimization context.
                        Determines whether any variable values should be fixed during the optimization
        :return: Tuple of (location of maximum, acquisition value at maximizer)
        """
        if context is None:
            context = dict()
        else:
            self._validate_context_parameters(context)
        context_manager = ContextManager(self.space, context)
        max_x, max_value = self._optimize(acquisition, context_manager)

        # Optimization might not match any encoding exactly
        # Rounding operation here finds the closest encoding
        rounded_max_x = np.concatenate(
            tuple(
                self.space.round(max_x[:, i * self.batch_size : (i + 1) * self.batch_size])
                for i in range(self.batch_size)
            ),
            axis=1,
        )

        return max_x, max_value



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

import numpy as np


class IModel:
    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predict mean and variance values for given points

        :param X: array of shape (n_points x n_inputs) of points to run prediction for
        :return: Tuple of mean and variance which are 2d arrays of shape (n_points x n_outputs)
        """
        raise NotImplementedError

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model

        :param X: new points
        :param Y: function values at new points X
        """
        raise NotImplementedError

    def optimize(self) -> None:
        """
        Optimize hyper-parameters of model
        """
        raise NotImplementedError

    @property
    def X(self):
        raise NotImplementedError

    @property
    def Y(self):
        raise NotImplementedError


class IDifferentiable:
    def get_prediction_gradients(self, X: np.ndarray) -> Tuple:
        """
        Computes and returns model gradients of mean and variance at given points

        :param X: points to compute gradients at
        :returns: Tuple of gradients of mean and variance.
        """
        raise NotImplementedError


class IJointlyDifferentiable:
    def predict_with_full_covariance(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 and n_points x n_points of the predictive
                 mean and variance at each input location
        """
        raise NotImplementedError

    def get_joint_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes and returns model gradients of mean and full covariance matrix at given points

        :param X: points to compute gradients at, nd array of shape (q, d)
        :return: Tuple with first item being gradient of the mean of shape (q) at X with respect to X (return shape is (q, q, d)).
                 The second item is the gradient of the full covariance matrix of shape (q, q) at X with respect to X
                 (return shape is (q, q, q, d)).
        """
        raise NotImplementedError


class IPriorHyperparameters:
    def generate_hyperparameters_samples(
        self, n_samples: int, n_burnin: int, subsample_interval: int, step_size: float, leapfrog_steps: int
    ) -> np.ndarray:
        """
        Generates the samples from the hyper-parameters of the model, and returns them.

        :param n_samples: Number of hyper-parameter samples
        :param n_burnin: Number of initial samples not used.
        :param subsample_interval: Interval of subsampling from HMC samples.
        :param step_size: Size of the gradient steps in the HMC sampler.
        :param leapfrog_steps: Number of gradient steps before each Metropolis Hasting step.
        """
        raise NotImplementedError

    def fix_model_hyperparameters(self, sample_hyperparameters: np.ndarray) -> None:
        """
        Fixes the model hyper-parameters to certain values (which can be taken from samples).

        :param sample_hyperparameters: np.ndarray whose rows contain each hyper-parameters set.
        """
        raise NotImplementedError


class IModelWithNoise:
    def predict_noiseless(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        For given points X, predict mean and variance of the output without observation noise.

        :param X: array of shape (n_points x n_inputs) of points to run prediction for
        :return: Tuple of mean and variance which are 2d arrays of shape (n_points x n_outputs)
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from emukit.core import ContinuousParameter, ParameterSpace


def sixhumpcamel_function():
    """
    Two-dimensional SixHumpCamel function, often used as an optimization benchmark.

    Based on: https://www.sfu.ca/~ssurjano/camel6.html

    .. math::
        f(\mathbf{x}) = \left(4-2.1x_1^2 = \frac{x_1^4}{3} \right)x_1^2 + x_1x_2 + (-4 +4x_2^2)x_2^2

    """

    parameter_space = ParameterSpace([ContinuousParameter("x1", -2, 2), ContinuousParameter("x2", -1, 1)])
    return _sixhumpcamel, parameter_space


def _sixhumpcamel(x):
    x1 = x[:, 0]
    x2 = x[:, 1]
    term1 = (4 - 2.1 * x1**2 + (x1**4) / 3) * x1**2
    term2 = x1 * x2
    term3 = (-4 + 4 * x2**2) * x2**2
    y = term1 + term2 + term3
    return y[:, None]



import itertools
import os
import sys
from copy import deepcopy
from importlib import reload
from typing import Callable, Dict, List, Tuple

import GPy
import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as la
import stan_utility
from GPy.inference.latent_function_inference.posterior import Posterior
from GPy.util.linalg import dpotri, dpotrs, jitchol, pdinv, tdot

# IF RUNNING ON CLUSTER AND RESTRICTED TO USE ONLY ONE CORE, UNCOMMENT THIS
# set environmental variable STAN_NUM_THREADS
# Use 4 cores per chain
# os.environ['STAN_NUM_THREADS'] = "1" # To make our lives easier on clusters


def remove_level_uncertainty(samples: np.ndarray, mean_same: bool = True) -> np.ndarray:
    """
    Removes level uncertainty in the samples. If only comparison observations are used, exact latent value doesn't matter.
    Thus in this case we make all samples have same maximum/mean
    :param samples: posterior samples
    :param mean_same: if True, all samples have same mean. If False, they have same maximum
    :return prosessed_samples: samples without level uncertainty
    """
    maxes = samples.max(axis=1)
    if mean_same:
        maxes = samples.mean(axis=1)
    reference = np.median(maxes)
    references = maxes - reference
    samples = samples - np.tile(references.reshape((-1, 1)), (1, samples.shape[1]))
    return samples


class StanPosterior(Posterior):
    """
    Posterior generated from the posterior samples
    :param samples: Stan samples
    :param X: input locations
    :param kern: Kernel to produce prior covariance matrices
    :param remove_levels: Boolean indicating if the level uncertainty should be removed from the samples
    :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param Y: kernel indices of each observation
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    """

    def __init__(
        self,
        samples: Dict,
        X: np.ndarray,
        kern: GPy.kern.Kern,
        noise: float,
        remove_levels: bool = True,
        y: List[Tuple[int, float]] = None,
        yc: List[List[Tuple[int, int]]] = None,
        get_logger: Callable = None,
        Y: np.ndarray = None,
    ):
        np.set_printoptions(precision=3, linewidth=500)
        self._K_chol, self._precision, self._woodbury_inv, self._woodbury_chol, self._woodbury_vector = (
            None,
            None,
            None,
            None,
            None,
        )
        samples["f"] = np.atleast_2d(samples["f"])
        if remove_levels:
            if get_logger is not None:
                get_logger().info("Removed level uncertainty")
            samples["f"] = remove_level_uncertainty(samples["f"])
        self._mean = np.array(np.mean(samples["f"].T, axis=1)).reshape((-1, 1))

        self._covariance = np.cov(samples["f"].T)

        self._prior_mean = 0

        self.samples = samples

        self.X = np.atleast_2d(X)

        self.Y = Y
        self.noise = noise
        self.kern = kern
        self._K = self.kern.K(self.X)

    def _get_mu_L(self, x_pred: np.ndarray, N: int = None, woodbury_inv: bool = False, with_index: int = None) -> Tuple:
        """
        Returns posterior mean and cholesky decomposition of the posterior samples

        :param x_pred: locations where the mean and posterior covariance are computed
        :param N: number of posterior samples
        :param woodbury_inv: boolean indicating whether the function should return woodbury_inv vector as well
        :param with_index: index of the specific posterior sample the function should return
        :return params: tuple containing the posterior means and choleskies of the covariances. Also woodbury inverses and woodbury choleskies if woodbury_inv is true
        """
        indices = np.arange(self.samples["f"].shape[0])
        if N is not None:
            indices = np.random.choice(indices, N)
        if with_index is not None:
            indices = np.array([with_index], dtype=int)
        N = len(indices)
        x_pred = np.atleast_2d(x_pred)
        f2_mu = np.empty((N, x_pred.shape[0]))
        f2_L = np.empty((N, x_pred.shape[0], x_pred.shape[0]))
        k_x1_x2 = self.kern.K(self.X, x_pred)
        k_x2_x2 = self.kern.K(x_pred)
        for ni, i in enumerate(indices):
            L_div_k_x1_x2 = la.solve_triangular(self.samples["L_K"][i, :, :], k_x1_x2, lower=True, overwrite_b=False)
            f2_mu[ni, :] = np.dot(L_div_k_x1_x2.T, self.samples["eta"][i, :])  # self.L_div_f[i,:])
            f2_cov = k_x2_x2 - np.dot(L_div_k_x1_x2.T, L_div_k_x1_x2)
            f2_L[ni, :, :] = jitchol(f2_cov)
        if woodbury_inv:
            w_inv = np.empty((N, self.X.shape[0], self.X.shape[0]))
            w_chol = np.empty(
                (
                    N,
                    self.X.shape[0],
                )
            )
            for ni, i in enumerate(indices):
                L_Kinv = la.inv(self.samples["L_K"][i, :, :])
                w_inv[ni, :, :] = L_Kinv.T @ L_Kinv
                w_chol[ni, :] = (L_Kinv.T @ self.samples["eta"][i, :, None])[
                    :, 0
                ]  # (Kinv @ self.samples['eta'][i,:, None])[:, 0] # (L_Kinv.T @ self.samples['eta'][i,:, None])[:, 0]  # self.L_div_f[i,:]
            return f2_mu, f2_L, w_inv, w_chol
        else:
            return f2_mu, f2_L

    def _raw_predict_raw(self, x_pred: np.ndarray, include_likelihood: bool = False, samples: int = 2000) -> np.ndarray:
        """
        Draw posterior predictive samples from the posterior distribution

        :param x_pred: locations at which the samples are wanted
        :param include_likelihood: boolean indicating if observation noise should be added to the samples
        :param samples: number of posterior predictive samples
        :return samples: posterior predictive samples
        """
        x_pred = np.atleast_2d(x_pred)
        f2_mu, f2_L = self._get_mu_L(x_pred, N=samples)
        f2 = f2_mu + np.matmul(f2_L, np.random.normal(size=(samples, x_pred.shape[0], 1)))[:, :, 0]
        if include_likelihood:
            f2 = f2 + self.noise * np.random.normal(size=(samples, x_pred.shape[0]))
        return f2



from itertools import cycle
from typing import List

import numpy as np

from .benchmark_result import BenchmarkResult

try:
    import matplotlib.pyplot as plt
except ImportError:
    ImportError("matplotlib needs to be installed in order to use BenchmarkPlot")


class BenchmarkPlot:
    """
    Creates a plot comparing the results from the different loops used during benchmarking
    """

    def __init__(
        self,
        benchmark_results: BenchmarkResult,
        loop_colours: List = None,
        loop_line_styles: List[str] = None,
        x_axis_metric_name: str = None,
        metrics_to_plot: List[str] = None,
    ):
        """
        :param benchmark_results: The output of a benchmark run
        :param loop_colours: Colours to use for each loop. Defaults to standard matplotlib colour palette
        :param loop_line_styles: Line style to use for each loop. Defaults to solid line style for all lines
        :param x_axis_metric_name: Which metric to use as the x axis in plots.
                                   None means it will be plotted against iteration number.
        :param metrics_to_plot: A list of metric names to plot. Defaults to all metrics apart from the one used as the
                                x axis.
        """
        self.benchmark_results = benchmark_results
        self.loop_names = benchmark_results.loop_names

        if loop_colours is None:
            self.loop_colours = _get_default_colours()
        else:
            self.loop_colours = loop_colours

        if loop_line_styles is None:
            self.loop_line_styles = ["-"]
        else:
            self.loop_line_styles = loop_line_styles

        if metrics_to_plot is None:
            self.metrics_to_plot = self.benchmark_results.metric_names
        else:
            for metric_name in metrics_to_plot:
                if metric_name not in self.benchmark_results.metric_names:
                    raise ValueError(metric_name + " not found in saved metrics from benchmark results.")
            self.metrics_to_plot = metrics_to_plot

        if x_axis_metric_name is not None:
            if x_axis_metric_name not in self.metrics_to_plot:
                raise ValueError("x_axis " + x_axis_metric_name + " is not a valid metric name")
            self.metrics_to_plot.remove(x_axis_metric_name)

        if x_axis_metric_name is None:
            self.x_label = "Iteration"
        else:
            self.x_label = x_axis_metric_name

        self.fig_handle = None
        self.x_axis = x_axis_metric_name

    def make_plot(self) -> None:
        """
        Make one plot for each metric measured, comparing the different loop results against each other
        """

        n_metrics = len(self.metrics_to_plot)
        self.fig_handle, _ = plt.subplots(n_metrics, 1)

        for i, metric_name in enumerate(self.metrics_to_plot):
            # Initialise plot
            plt.subplot(n_metrics, 1, i + 1)
            plt.title(metric_name)

            colours = cycle(self.loop_colours)
            line_styles = cycle(self.loop_line_styles)
            min_x = np.inf
            max_x = -np.inf

            for j, loop_name in enumerate(self.loop_names):
                # Get all results for this metric
                metric = self.benchmark_results.extract_metric_as_array(loop_name, metric_name)

                # Get plot options
                colour = next(colours)
                line_style = next(line_styles)

                # Get data to plot
                mean, std = _get_metric_stats(metric)

                if self.x_axis is not None:
                    x = np.mean(self.benchmark_results.extract_metric_as_array(loop_name, self.x_axis), axis=0)
                else:
                    x = np.arange(0, mean.shape[0])

                # Save min/max of data to set the axis limits later
                min_x = np.min([np.min(x), min_x])
                max_x = np.max([np.max(x), max_x])

                # Plot
                plt.plot(x, mean, color=colour, linestyle=line_style)
                plt.xlabel(self.x_label)
                plt.fill_between(x, mean - std, mean + std, alpha=0.2, color=colour)

            # Make legend
            plt.legend(self.loop_names)
            plt.tight_layout()

            plt.xlim(min_x, max_x)

    def save_plot(self, file_name: str) -> None:
        """
        Save plot to file

        :param file_name:
        """
        if self.fig_handle is None:
            raise ValueError("Please call make_plots method before saving to file")

        with open(file_name) as file:
            self.fig_handle.savefig(file)


def _get_metric_stats(metric):
    return metric.mean(axis=0), metric.std(axis=0)


def _get_default_colours():
    return plt.rcParams["axes.prop_cycle"].by_key()["color"]



from typing import Tuple

import numpy as np

from ...core import ContinuousParameter, InformationSourceParameter, ParameterSpace
from ...core.loop.user_function import MultiSourceFunctionWrapper


def multi_fidelity_park_function() -> Tuple[MultiSourceFunctionWrapper, ParameterSpace]:
    r"""

    High fidelity is given by:

    .. math::
        f_{high}(\mathbf{x}) = \frac{x_1}{2}\left[ \sqrt{1 + \left( x_2 + x_3^2 \right) \frac{x_4}{x_1^2}} - 1\right] +
        \left( x_1 + 3 x_4 \right) \exp \left[ 1 + \sin x_3 \right]

    Low fidelity is given by:

    .. math::
        f_{low}(\mathbf{x}) = \left[ 1 + \frac{\sin x_1}{10} \right] f_{high}(\mathbf{x}) -
        2x_1 + x_2 ^ 2 + x_3 ^2 + 0.5

    The input domain is given by:

    .. math::
        \mathbf{x}_i \in (0, 1)

    Reference: https://www.sfu.ca/~ssurjano/park91a.html
    """

    def park_high(x):
        x1 = x[:, 0]
        x2 = x[:, 1]
        x3 = x[:, 2]
        x4 = x[:, 3]

        tmp = 1 + (x2 + x3) * (x4 / x1**2)
        return ((x1 / 2) * (np.sqrt(tmp) - 1) + (x1 + 3 * x4) * np.exp(1 + np.sin(x3)))[:, None]

    def park_low(x):
        f_high = park_high(x)

        x1 = x[:, 0]
        x2 = x[:, 1]
        x3 = x[:, 2]

        return ((1 + np.sin(x1) / 10) * f_high.flatten() - 2 * x1 + x2**2 + x3**2 + 0.5)[:, None]

    space = ParameterSpace(
        [
            ContinuousParameter("x1", 0.0, 1.0),
            ContinuousParameter("x2", 0.0, 1.0),
            ContinuousParameter("x3", 0.0, 1.0),
            ContinuousParameter("x4", 0.0, 1.0),
            InformationSourceParameter(2),
        ]
    )

    return MultiSourceFunctionWrapper([park_low, park_high]), space



import numpy as np

from emukit.core.initial_designs.latin_design import LatinDesign
from emukit.core.loop import FixedIterationsStoppingCondition, LoopState, UserFunctionWrapper
from emukit.core.parameter_space import ParameterSpace
from emukit.examples.fabolas import FabolasLoop


def fmin_fabolas(
    func,
    space: ParameterSpace,
    s_min: float,
    s_max: float,
    n_iters: int,
    n_init: int = 20,
    marginalize_hypers: bool = True,
) -> LoopState:
    """
    Simple interface for Fabolas which optimizes the hyperparameters of machine learning algorithms
    by reasoning across training data set subsets. For further details see:

    Fast Bayesian hyperparameter optimization on large datasets
    A. Klein and S. Falkner and S. Bartels and P. Hennig and F. Hutter
    Electronic Journal of Statistics (2017)

    :param func: objective function which gets a hyperparameter configuration x and training dataset size s as input,
    and return the validation error and the runtime after training x on s datapoints.
    :param space: input space
    :param s_min: minimum training dataset size (linear scale)
    :param s_max: maximum training dataset size (linear scale)
    :param n_iters: number of iterations
    :param n_init: number of initial design points (needs to be smaller than num_iters)
    :param marginalize_hypers: determines whether to use a MAP estimate or to marginalize over the GP hyperparameters

    :return: LoopState with all evaluated data points
    """
    initial_design = LatinDesign(space)

    grid = initial_design.get_samples(n_init)
    X_init = np.zeros([n_init, grid.shape[1] + 1])
    Y_init = np.zeros([n_init, 1])
    cost_init = np.zeros([n_init])

    subsets = np.array([s_max // 2**i for i in range(2, 10)])[::-1]
    idx = np.where(subsets < s_min)[0]

    subsets[idx] = s_min

    for it in range(n_init):
        func_val, cost = func(x=grid[it], s=subsets[it % len(subsets)])

        X_init[it] = np.concatenate((grid[it], np.array([subsets[it % len(subsets)]])))
        Y_init[it] = func_val
        cost_init[it] = cost

    def wrapper(x):
        y, c = func(x[0, :-1], x[0, -1])

        return np.array([[y]]), np.array([[c]])

    loop = FabolasLoop(
        X_init=X_init,
        Y_init=Y_init,
        cost_init=cost_init,
        space=space,
        s_min=s_min,
        s_max=s_max,
        marginalize_hypers=marginalize_hypers,
    )
    loop.run_loop(
        user_function=UserFunctionWrapper(wrapper, extra_output_names=["cost"]),
        stopping_condition=FixedIterationsStoppingCondition(n_iters - n_init),
    )

    return loop.loop_state



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from .. import ParameterSpace
from .base import InitialDesignBase


class RandomDesign(InitialDesignBase):
    """
    Random experiment design.
    Uniform random values for all variables within the given bounds.
    """

    def __init__(self, parameter_space: ParameterSpace) -> None:
        """
        :param parameter_space: The parameter space to generate design for.
        """
        super(RandomDesign, self).__init__(parameter_space)

    def get_samples(self, point_count: int) -> np.ndarray:
        """
        Generates requested amount of points.

        :param point_count: Number of points required.
        :return: A numpy array of generated samples, shape (point_count x space_dim)
        """
        return self.parameter_space.sample_uniform(point_count)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .acquisition import Acquisition  # noqa: F401
from .acquisition_per_cost import acquisition_per_expected_cost  # noqa: F401
from .integrated_acquisition import IntegratedHyperParameterAcquisition  # noqa: F401



import logging
import math
from functools import partial
from typing import Dict, Tuple

import evalset.test_funcs
import GPy
import numpy as np

import emukit
from emukit.bayesian_optimization.loops import BayesianOptimizationLoop
from emukit.core import ContinuousParameter, ParameterSpace
from emukit.core.optimization import GradientAcquisitionOptimizer

from . import ComparisonGP, ComparisonGPEmukitWrapper, util
from .acquisitions import (
    AcquisitionFunction,
    EmukitAcquisitionFunctionWrapper,
    SequentialGradientAcquisitionOptimizer,
    ThompsonSampling,
)


def create_bayesian_optimization_loop(
    gpy_model: ComparisonGP, lims: np.array, batch_size: int, acquisition: AcquisitionFunction
) -> BayesianOptimizationLoop:
    """
    Creates Bayesian optimization loop for Bayesian neural network or random forest models.
    :param gpy_model: the GPy model used in optimization
    :param lims: Optimization limits for the inputs
    :param batch_size: number of observations used in batch
    :param acquisition: acquisition function used in the bayesian optimization
    :return: emukit BO loop
    """

    # Create model
    model = ComparisonGPEmukitWrapper(gpy_model, batch_size)

    # Create acquisition
    emukit_acquisition = EmukitAcquisitionFunctionWrapper(model, acquisition)

    if type(emukit_acquisition.acquisitionFunction) is ThompsonSampling:
        parameter_space = []
        for j in range(len(lims)):
            parameter_space += [ContinuousParameter("x{}".format(j), lims[j][0], lims[j][1])]
        parameter_space = ParameterSpace(parameter_space)
        acquisition_optimizer = SequentialGradientAcquisitionOptimizer(parameter_space, batch_size)
    else:
        parameter_space = []
        for k in range(batch_size):
            for j in range(len(lims)):
                parameter_space += [ContinuousParameter("x{}{}".format(k, j), lims[j][0], lims[j][1])]
        parameter_space = ParameterSpace(parameter_space)
        acquisition_optimizer = GradientAcquisitionOptimizer(parameter_space)

    bo = BayesianOptimizationLoop(model=model, space=parameter_space, acquisition=emukit_acquisition)
    return BayesianOptimizationLoop(
        model=model, space=parameter_space, acquisition=emukit_acquisition, acquisition_optimizer=acquisition_optimizer
    )


class BayesianOptimization:
    """
    This is a simple implementation of Bayesian optimization needed for preforming preferential batch Bayesian optimization

    The class structure is heavily inspired by https://github.com/oxfordcontrol/Bayesian-Optimization/blob/GPy-based/methods/bo.py

    :param options: a dictionary containing otions needed by the BO loop.
                    The dictionary must contain values for 'objective' (type: TestFunction),
                    'inference' (type: CompartisonGP), 'acquisition' (type: AcquisitionFunction),
                    'kernel' (type: GPy.kern.Kern) and 'batch_size' (type:int)
    """

    def __init__(self, options: Dict):
        self.options = options.copy()

        self.kernel = options["kernel"].copy()  # GPy kernel used for the optimization
        self.inference = options["inference"]  # Inference method used to approximate the posterior
        self.acquisition = options["acquisition"]  # The acquisition function used to get new points
        self.batch_size = options["batch_size"]  # Number of observations in a batch
        self.max_num_observations = options["max_num_observations"]  # Maximum number of observations allowed

        # Number of iterations the BO loop is ran can be computed from the max number of observations and batch size
        self.iterations = math.ceil(float(self.max_num_observations) / float(self.batch_size))

        # Optional parameters that have defaults
        self.noise = options.get("noise", 1e-6)  # Noise std the likelihood uses for observations

        # If we use comparative observations from the objective function before starting
        self.use_comparative_observations_in_init = options.get("use_comparative_observations_in_init", False)
        # If we use direct observations from the objective function before starting
        self.use_direct_observations_in_init = options.get("use_direct_observations_in_init", False)
        # Boolean indicating if the locations for the initial draws are chosen at random or selected uniformly from a grid
        self.random = options.get("random", False)

        self.log_file = options.get("log_file", None)  # Log file used to store all print outputs during the BO

        # Configure logger
        util.configure_logger(log_file=self.log_file)

    def get_logger(self) -> logging.Logger:
        """
        Get the logging function used to save the print outputs
        :return: a logger where the prints are directed
        """
        return logging.getLogger(self.log_file)

    def bayesian_optimization(self, objective: evalset.test_funcs.TestFunction) -> Tuple[np.ndarray, np.ndarray]:
        """
        This function implements the main loop of Bayesian Optimization,

        The function takes the objective

        The function returns the final set of points X_all, the respective
        batch feedbacks yc_all

        :param objective: The black box function to be optimized
        :return: The locations the function has been evaluated so far and the comparison outcomes
        """

        # SET UP THE GP MODEL #
        bounds = objective.bounds
        dim = len(bounds)

        lik = GPy.likelihoods.Gaussian()
        lik.variance.constrain_fixed(self.noise**2, warning=False)
        noise = self.noise

        X0 = np.empty((0, dim))
        y = []
        yc = []

        def objective_modifier(x, f=None, batch_size=1):
            return np.concatenate(
                tuple(
                    f(x[:, i * batch_size : (i + 1) * batch_size]).reshape((-1, 1))
                    for i in range(x.shape[1] // batch_size)
                ),
                axis=1,
            )

        # Initial observations:
        if self.use_comparative_observations_in_init:
            if self.random:
                X0 = util.random_sample(bounds, 2**dim)
            else:
                X0 = util.grid_sample(dim)
            yc = util.give_comparisons(objective.f, X0)
        if self.use_direct_observations_in_init:
            if self.random:
                Xn = util.random_sample(bounds, 2**dim)
            else:
                Xn = util.grid_sample(dim)
            yn = objective.f(Xn).reshape((-1, 1))
            y = [(X0.shape[0] + i, yi) for i, yi in enumerate(yn)]
            X0 = np.concatenate((X0, Xn), axis=0)

        if not self.use_comparative_observations_in_init and not self.use_direct_observations_in_init:
            m = self.inference(
                util.static_sample(bounds),
                [(i, yi) for i, yi in enumerate(np.array([[0], [0]]))],
                yc,
                self.kernel.copy(),
                lik,
                get_logger=self.get_logger,
            )
        else:
            m = self.inference(X0, y, yc, self.kernel.copy(), lik, get_logger=self.get_logger)

        # CREATE BO LOOP #
        bo_loop = create_bayesian_optimization_loop(m, bounds, self.batch_size, self.acquisition)

        # RUN THE LOOP #
        bo_loop.run_loop(partial(objective_modifier, f=objective.f, batch_size=self.batch_size), self.iterations)
        return m.X, m.yc



"""The RBF kernel embeddings."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Union

import numpy as np
from scipy.special import erf

from ...quadrature.interfaces.standard_kernels import IRBF
from ..kernels import GaussianEmbedding, LebesgueEmbedding, QuadratureKernel
from ..measures import GaussianMeasure, IntegrationMeasure, LebesgueMeasure


class QuadratureRBF(QuadratureKernel):
    r"""Base class for an RBF kernel augmented with integrability.

    .. math::
        k(x, x') = \sigma^2 e^{-\frac{1}{2}\sum_{i=1}^{d}r_i^2},

    where :math:`d` is the input dimensionality,
    :math:`r_i = \frac{x_i-x_i'}{\lambda_i}` is the scaled vector difference of dimension :math:`i`,
    :math:`\lambda_i` is the :math:`i` th element of the ``lengthscales`` property
    and :math:`\sigma^2` is the ``variance`` property.

    .. note::
        This class is compatible with the standard kernel :class:`IRBF`.
        Each child of this class implements an embedding w.r.t. a specific integration measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IRBF`
       * :class:`emukit.quadrature.kernels.QuadratureKernel`

    :param rbf_kernel: The standard EmuKit rbf-kernel.
    :param measure: The integration measure. ``None`` implies the standard Lebesgue measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(
        self,
        rbf_kernel: IRBF,
        measure: IntegrationMeasure,
        variable_names: str,
    ) -> None:
        super().__init__(kern=rbf_kernel, measure=measure, variable_names=variable_names)

    @property
    def lengthscales(self) -> np.ndarray:
        r"""The lengthscales :math:`\lambda` of the kernel."""
        return self.kern.lengthscales

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        return self.kern.variance

    # rbf-kernel specific helper
    def _scaled_vector_diff(
        self, v1: np.ndarray, v2: np.ndarray, scales: Union[float, np.ndarray] = None
    ) -> np.ndarray:
        r"""Scaled element-wise vector difference between vectors v1 and v2.

        .. math::
            \frac{v_1 - v_2}{\ell \sqrt{2}}

        where :math:`\ell` is the ``scales`` parameter.

        :param v1: First vector.
        :param v2: Second vector, must have same second dimensions as v1.
        :param scales: The scales, default is the lengthscales of the kernel.
        :return: Scaled difference between v1 and v2, same shape as v1 and v2.
        """
        if scales is None:
            scales = self.lengthscales
        return (v1 - v2) / (scales * np.sqrt(2))


class QuadratureRBFLebesgueMeasure(QuadratureRBF, LebesgueEmbedding):
    """An RBF kernel augmented with integrability w.r.t. the standard Lebesgue measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IRBF`
       * :class:`emukit.quadrature.kernels.QuadratureRBF`
       * :class:`emukit.quadrature.measures.LebesgueMeasure`

    :param rbf_kernel: The standard EmuKit rbf-kernel.
    :param measure: The Lebesgue measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(self, rbf_kernel: IRBF, measure: LebesgueMeasure, variable_names: str = "") -> None:
        super().__init__(rbf_kernel=rbf_kernel, measure=measure, variable_names=variable_names)

    def qK(self, x2: np.ndarray) -> np.ndarray:
        lb = self.measure.domain.lower_bounds[None, :]
        ub = self.measure.domain.upper_bounds[None, :]
        erf_lo = erf(self._scaled_vector_diff(lb, x2))
        erf_up = erf(self._scaled_vector_diff(ub, x2))
        kernel_mean = (np.sqrt(np.pi / 2.0) * self.lengthscales * (erf_up - erf_lo)).prod(axis=1)
        return (self.variance * self.measure.density) * kernel_mean.reshape(1, -1)

    def qKq(self) -> float:
        lb = self.measure.domain.lower_bounds[None, :]
        ub = self.measure.domain.upper_bounds[None, :]
        diff_bounds_scaled = self._scaled_vector_diff(ub, lb)
        exp_term = (np.exp(-(diff_bounds_scaled**2)) - 1.0) / np.sqrt(np.pi)
        erf_term = erf(diff_bounds_scaled) * diff_bounds_scaled
        qKq = ((2 * np.sqrt(np.pi) * self.lengthscales**2) * (exp_term + erf_term)).prod()
        return (self.variance * self.measure.density**2) * float(qKq)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        lb = self.measure.domain.lower_bounds[None, :]
        ub = self.measure.domain.upper_bounds[None, :]
        exp_lo = np.exp(-self._scaled_vector_diff(x2, lb) ** 2)
        exp_up = np.exp(-self._scaled_vector_diff(x2, ub) ** 2)
        erf_lo = erf(self._scaled_vector_diff(lb, x2))
        erf_up = erf(self._scaled_vector_diff(ub, x2))
        fraction = ((exp_lo - exp_up) / (self.lengthscales * np.sqrt(np.pi / 2.0) * (erf_up - erf_lo))).T
        return self.qK(x2) * fraction


class QuadratureRBFGaussianMeasure(QuadratureRBF, GaussianEmbedding):
    """An RBF kernel augmented with integrability w.r.t. a Gaussian measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IRBF`
       * :class:`emukit.quadrature.kernels.QuadratureRBF`
       * :class:`emukit.quadrature.measures.GaussianMeasure`

    :param rbf_kernel: The standard EmuKit rbf-kernel.
    :param measure: A Gaussian measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(self, rbf_kernel: IRBF, measure: GaussianMeasure, variable_names: str = "") -> None:
        super().__init__(rbf_kernel=rbf_kernel, measure=measure, variable_names=variable_names)

    def qK(self, x2: np.ndarray, scale_factor: float = 1.0) -> np.ndarray:
        lengthscales = scale_factor * self.lengthscales
        sigma2 = self.measure.variance
        mu = self.measure.mean
        factor = np.sqrt(lengthscales**2 / (lengthscales**2 + sigma2)).prod()
        scaled_norm_sq = np.power(self._scaled_vector_diff(x2, mu, np.sqrt(lengthscales**2 + sigma2)), 2).sum(axis=1)
        return (self.variance * factor) * np.exp(-scaled_norm_sq).reshape(1, -1)

    def qKq(self) -> float:
        lengthscales = self.lengthscales
        qKq = np.sqrt(lengthscales**2 / (lengthscales**2 + 2 * self.measure.variance)).prod()
        return self.variance * float(qKq)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        scaled_diff = (x2 - self.measure.mean) / (self.lengthscales**2 + self.measure.variance)
        return -self.qK(x2) * scaled_diff.T



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import List, Optional, Sequence, Tuple

import numpy as np

from .. import (
    CategoricalParameter,
    ContinuousParameter,
    DiscreteParameter,
    OneHotEncoding,
    OrdinalEncoding,
    Parameter,
    ParameterSpace,
)
from ..acquisition import Acquisition
from .acquisition_optimizer import AcquisitionOptimizerBase
from .context_manager import ContextManager

_log = logging.getLogger(__name__)


class LocalSearchAcquisitionOptimizer(AcquisitionOptimizerBase):
    """Optimizes the acquisition function by multiple local searches starting at random points.
    Each local optimization iteratively evaluates the one-exchange neighbourhoods.
    Can be used for discrete and continuous acquisition functions.

    This kind of optimization is also known as Variable Neighbourhood Search
    (e.g. see https://en.wikipedia.org/wiki/Variable_neighborhood_search).
    Neighbourhood definitions and default parameters are based on the search used
    in SMAC [1].

    .. warning:: The local search heuristic here currently differs to SMAC [1].
                 The neighbourhood of a point is evaluated completely,
                 the search continues at the best neighbour (best improvement heuristic).
                 SMAC iteratively samples neighbours and continues at the first which
                 is better than the current (first improvement heuristic).
                 Therefore this implementation is time consuming for large neighbourhoods
                 (e.g. parameters with hundreds of categories).

    One-exchange neighbourhood is defined for the following parameter types:
      :Categorical parameter with one-hot encoding: All other categories
      :Categorical parameter with ordinal encoding: Only preceeding and following categories
      :Continuous parameter: Gaussian samples (default: 4) around current value. Standard deviation (default: 0.2) is
                             scaled by parameter value range.
      :Discrete parameter: Preceeding and following discrete values.

    .. [1] Hutter, Frank, Holger H. Hoos, and Kevin Leyton-Brown.
           "Sequential model-based optimization for general algorithm configuration."
           International Conference on Learning and Intelligent Optimization.
           Springer, Berlin, Heidelberg, 2011.
    """

    def __init__(
        self,
        space: ParameterSpace,
        num_steps: int = 10,
        num_init_points: int = 5,
        std_dev: float = 0.02,
        num_continuous: int = 4,
    ) -> None:
        """
        :param space: The parameter space spanning the search problem.
        :param num_steps: Maximum number of steps to follow from each start point.
        :param num_init_points: Number of initial sampled points where the local search starts.
        :param std_dev: Neighbourhood sampling standard deviation of continuous parameters.
        :param num_continuous: Number of sampled neighbourhoods per continuous parameter.
        """
        super().__init__(space)
        self.num_steps = num_steps
        self.num_init_points = num_init_points
        self.std_dev = std_dev
        self.num_continuous = num_continuous

    def _neighbours_per_parameter(self, all_features: np.ndarray, parameters: Sequence[Parameter]) -> List[np.ndarray]:
        """Generates parameter encodings for one-exchange neighbours of
            parameters encoded in parameter feature vector

        :param all_features: The encoded parameter point (1d-array)
        :return: List of numpy arrays. Each array contains all one-exchange encodings of a parameter
        """
        neighbours = []
        current_feature = 0
        for parameter in parameters:
            features = parameter.round(
                all_features[current_feature : (current_feature + parameter.dimension)].reshape(1, -1)
            ).ravel()
            if isinstance(parameter, CategoricalParameter):
                if isinstance(parameter.encoding, OrdinalEncoding):
                    left_right = np.unique(
                        [parameter.encoding.round_row(features - 1), parameter.encoding.round_row(features + 1)]
                    )
                    neighbours.append(left_right[left_right != features].reshape(-1, 1))
                elif isinstance(parameter.encoding, OneHotEncoding):
                    # All categories apart from current one are valid neighbours with one hot encoding
                    neighbours.append(parameter.encodings[(parameter.encodings != features).any(axis=1)])
                else:
                    raise TypeError("{} not a supported parameter encoding.".format(type(parameter.encoding)))
            elif isinstance(parameter, DiscreteParameter):
                # Find current position in domain while being robust to numerical precision problems
                current_index = np.argmin(np.abs(np.subtract(parameter.domain, features.item())))
                this_neighbours = []
                if current_index > 0:
                    this_neighbours.append([parameter.domain[current_index - 1]])
                if current_index < len(parameter.domain) - 1:
                    this_neighbours.append([parameter.domain[current_index + 1]])
                neighbours.append(np.asarray(this_neighbours).reshape(-1, 1))
            elif isinstance(parameter, ContinuousParameter):
                samples, param_range = [], parameter.max - parameter.min
                while len(samples) < self.num_continuous:
                    sample = np.random.normal(features.item(), self.std_dev * param_range, (1, 1))
                    if parameter.min <= sample <= parameter.max:
                        samples.append(sample)
                neighbours.append(np.vstack(samples))
            else:
                raise TypeError("{} not a supported parameter type.".format(type(parameter)))
            current_feature += parameter.dimension
        return neighbours

    def _neighbours(self, all_features: np.ndarray, parameters: Sequence[Parameter]) -> np.ndarray:
        """Generates one-exchange neighbours of encoded parameter point.

        :param all_features: The encoded parameter point (1d-array)
        :return: All one-exchange neighbours as 2d-array (neighbours, features)
        """
        neighbours_per_param = self._neighbours_per_parameter(all_features, parameters)
        num_neighbours = sum(param.shape[0] for param in neighbours_per_param)
        neighbours = np.full((num_neighbours, all_features.shape[0]), all_features)
        current_neighbour, current_feature = 0, 0
        for this_neighbours in neighbours_per_param:
            next_neighbour = current_neighbour + this_neighbours.shape[0]
            next_feature = current_feature + this_neighbours.shape[1]
            neighbours[current_neighbour:next_neighbour, current_feature:next_feature] = this_neighbours
            current_neighbour, current_feature = next_neighbour, next_feature
        return neighbours

    def _one_local_search(
        self, acquisition: Acquisition, x: np.ndarray, context_manager: Optional[ContextManager] = None
    ):
        """Local maximum search on acquisition starting at a single point.

        :param acquisition: The acquisition the maximum is searched of.
        :param x: The initial point.
        :return: Tuple of (maximum point as 1d-array, value of acquisition at this point)
        """
        incumbent_value = acquisition.evaluate(x.reshape(1, -1)).item()
        _log.debug("Start local search with acquisition={:.4f} at {}".format(incumbent_value, str(x)))
        for step in range(self.num_steps):
            neighbours = self._neighbours(x, context_manager.contextfree_space.parameters)
            neighbours_with_context = context_manager.expand_vector(neighbours)
            acquisition_values = acquisition.evaluate(neighbours_with_context)
            max_index = np.argmax(acquisition_values)
            max_neighbour = neighbours[max_index]
            max_value = acquisition_values[max_index].item()
            if max_value < incumbent_value:
                _log.debug(
                    "End after {} steps at maximum of acquisition={:.4f} at {}".format(step, incumbent_value, str(x))
                )
                return x, incumbent_value
            else:
                incumbent_value = max_value
                x = max_neighbour
        _log.debug("End at step limit with acquisition={:.4f} at {}".format(incumbent_value, str(x)))
        return x, incumbent_value

    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager) -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of abstract method.

        See AcquisitionOptimizerBase._optimizer for parameter descriptions.
        See class docstring for implementation details.
        """
        X_init = context_manager.contextfree_space.sample_uniform(self.num_init_points)
        X_max = np.empty_like(X_init)
        acq_max = np.empty((self.num_init_points, 1))
        _log.info("Starting local optimization of acquisition function {}".format(type(acquisition)))
        for sample in range(self.num_init_points):  # this loop could be parallelized
            X_max[sample], acq_max[sample] = self._one_local_search(acquisition, X_init[sample], context_manager)
        max_index = np.argmax(acq_max)
        X_max_with_context = context_manager.expand_vector(X_max)
        return X_max_with_context[[max_index]], acq_max[[max_index]]



"""The Brownian motion kernel embeddings."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Union

import numpy as np

from ...quadrature.interfaces.standard_kernels import IBrownian, IProductBrownian
from ..kernels import LebesgueEmbedding, QuadratureKernel, QuadratureProductKernel
from ..measures import IntegrationMeasure, LebesgueMeasure


class QuadratureBrownian(QuadratureKernel):
    r"""Base class for a Brownian motion kernel augmented with integrability.

    .. math::
        k(x, x') = \sigma^2 \operatorname{min}(x, x')\quad\text{with}\quad x, x' \geq 0,

    where :math:`\sigma^2` is the ``variance`` property.

    .. note::
        This class is compatible with the standard kernel :class:`IBrownian`.
        Each child of this class implements an embedding w.r.t. a specific integration measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IBrownian`
       * :class:`emukit.quadrature.kernels.QuadratureKernel`

    :param brownian_kernel: The standard EmuKit Brownian motion kernel.
    :param measure: The integration measure.
    :param variable_names: The (variable) name(s) of the integral.

    :raises ValueError: If ``measure`` has wrong dimensionality.
    :raises ValueError: If the reasonable box of the measure allows for negative values.

    """

    def __init__(
        self,
        brownian_kernel: IBrownian,
        measure: IntegrationMeasure,
        variable_names: str,
    ) -> None:

        if measure.input_dim != 1:
            raise ValueError(
                "Integration measure for Brownian motion kernel must be 1-dimensional. Current dimesnion is ({}).".format(
                    measure.input_dim
                )
            )

        super().__init__(kern=brownian_kernel, measure=measure, variable_names=variable_names)

        if any(self.reasonable_box.lower_bounds < 0):
            raise ValueError(
                "The domain defined by the reasonable box seems to allow negative values. "
                "Brownian motion is only defined for positive input values."
            )

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        return self.kern.variance


class QuadratureBrownianLebesgueMeasure(QuadratureBrownian, LebesgueEmbedding):
    """A Brownian motion kernel augmented with integrability w.r.t. the standard Lebesgue measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IBrownian`
       * :class:`emukit.quadrature.kernels.QuadratureBrownian`
       * :class:`emukit.quadrature.measures.LebesgueMeasure`

    :param brownian_kernel: The standard EmuKit Brownian motion kernel.
    :param measure: The Lebesgue measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(self, brownian_kernel: IBrownian, measure: LebesgueMeasure, variable_names: str = "") -> None:
        super().__init__(brownian_kernel=brownian_kernel, measure=measure, variable_names=variable_names)

    def qK(self, x2: np.ndarray) -> np.ndarray:
        lb = self.measure.domain.lower_bounds[None, :]
        ub = self.measure.domain.upper_bounds[None, :]
        kernel_mean = ub * x2 - 0.5 * x2**2 - 0.5 * lb**2
        return (self.variance * self.measure.density) * kernel_mean.T

    def qKq(self) -> float:
        lb = self.measure.domain.lower_bounds[None, :]
        ub = self.measure.domain.upper_bounds[None, :]
        qKq = 0.5 * ub * (ub**2 - lb**2) - (ub**3 - lb**3) / 6 - 0.5 * lb**2 * (ub - lb)
        return (self.measure.density**2 * self.variance) * float(qKq)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        ub = self.measure.domain.upper_bounds[None, :]
        return (self.measure.density * self.variance) * (ub - x2).T


class QuadratureProductBrownian(QuadratureProductKernel):
    r"""Base class for a product Brownian kernel augmented with integrability.

    The kernel is of the form :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = \operatorname{min}(x_i-c, x_i'-c)\quad\text{with}\quad x_i, x_i' \geq c,

    :math:`d` is the input dimensionality,
    :math:`\sigma^2` is the ``variance`` property
    and :math:`c` is the ``offset`` property.

    .. note::
        This class is compatible with the standard kernel :class:`IProductBrownian`.
        Each subclass of this class implements an embedding w.r.t. a specific integration measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IProductBrownian`
       * :class:`emukit.quadrature.kernels.QuadratureProductKernel`

    :param brownian_kernel: The standard EmuKit product Brownian kernel.
    :param measure: The integration measure.
    :param variable_names: The (variable) name(s) of the integral.

    :raises ValueError: If the reasonable box of the measure allows for values smaller than the offset :math:`c`.

    """

    def __init__(
        self,
        brownian_kernel: IProductBrownian,
        measure: IntegrationMeasure,
        variable_names: str,
    ) -> None:
        super().__init__(kern=brownian_kernel, measure=measure, variable_names=variable_names)

        lower_bounds_x = self.reasonable_box.lower_bounds
        if any(lower_bounds_x < self.offset):
            raise ValueError(
                f"The domain defined by the reasonable box seems allow to for values smaller than the offset "
                f"({self.offset}). Brownian motion is only defined for input values larger than the offset."
            )

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        return self.kern.variance

    @property
    def offset(self) -> float:
        r"""The offset :math:`c` of the kernel."""
        return self.kern.offset


class QuadratureProductBrownianLebesgueMeasure(QuadratureProductBrownian, LebesgueEmbedding):
    """A product Brownian kernel augmented with integrability w.r.t. the standard Lebesgue measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IProductBrownian`
       * :class:`emukit.quadrature.kernels.QuadratureProductBrownian`
       * :class:`emukit.quadrature.measures.LebesgueMeasure`

    :param brownian_kernel: The standard EmuKit product Brownian kernel.
    :param measure: The Lebesgue measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(self, brownian_kernel: IProductBrownian, measure: LebesgueMeasure, variable_names: str = "") -> None:
        super().__init__(brownian_kernel=brownian_kernel, measure=measure, variable_names=variable_names)

    def _scale(self, z: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return self.variance * z

    def _get_univariate_parameters(self, dim: int) -> dict:
        return {
            "domain": self.measure.domain.bounds[dim],
            "offset": self.offset,
            "normalize": self.measure.is_normalized,
        }

    def _qK_1d(self, x: np.ndarray, **parameters) -> np.ndarray:
        a, b = parameters["domain"]
        offset = parameters["offset"]
        normalization = 1 / (b - a) if parameters["normalize"] else 1.0
        kernel_mean = b * x - 0.5 * x**2 - 0.5 * a**2
        return (kernel_mean.T - offset * (b - a)) * normalization

    def _qKq_1d(self, **parameters) -> float:
        a, b = parameters["domain"]
        offset = parameters["offset"]
        normalization = 1 / (b - a) if parameters["normalize"] else 1.0
        qKq = 0.5 * b * (b**2 - a**2) - (b**3 - a**3) / 6 - 0.5 * a**2 * (b - a)
        return (float(qKq) - offset * (b - a) ** 2) * normalization**2

    def _dqK_dx_1d(self, x: np.ndarray, **parameters) -> np.ndarray:
        a, b = parameters["domain"]
        normalization = 1 / (b - a) if parameters["normalize"] else 1.0
        return (b - x).T * normalization



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import abc

import numpy as np

from ..acquisition import Acquisition
from ..interfaces import IModel
from ..optimization.acquisition_optimizer import AcquisitionOptimizerBase
from ..optimization.context_manager import ContextManager
from ..parameter_space import ParameterSpace
from .loop_state import LoopState


class CandidatePointCalculator(abc.ABC):
    """Computes the next point(s) for function evaluation"""

    @abc.abstractmethod
    def compute_next_points(self, loop_state: LoopState, context: dict = None) -> np.ndarray:
        """
        :param loop_state: Object that contains current state of the loop
        :param context: Contains variables to fix through optimization of acquisition function. The dictionary key is
                        the parameter name and the value is the value to fix the parameter to.
        :return: (n_points x n_dims) array of next inputs to evaluate the function at
        """
        pass


class SequentialPointCalculator(CandidatePointCalculator):
    """This candidate point calculator chooses one candidate point at a time"""

    def __init__(self, acquisition: Acquisition, acquisition_optimizer: AcquisitionOptimizerBase) -> None:
        """
        :param acquisition: Acquisition function to maximise
        :param acquisition_optimizer: Optimizer of acquisition function
        """
        self.acquisition = acquisition
        self.acquisition_optimizer = acquisition_optimizer

    def compute_next_points(self, loop_state: LoopState, context: dict = None) -> np.ndarray:
        """
        Computes point(s) to evaluate next

        :param loop_state: Object that contains current state of the loop
        :param context: Contains variables to fix through optimization of acquisition function. The dictionary key is
                        the parameter name and the value is the value to fix the parameter to.
        :return: List of function inputs to evaluate the function at next
        """
        self.acquisition.update_parameters()
        x, _ = self.acquisition_optimizer.optimize(self.acquisition, context)
        return x


class GreedyBatchPointCalculator(CandidatePointCalculator):
    """
    Batch point calculator. This point calculator calculates the first point in the batch then adds this as a fake
    observation in the model with a Y value equal to the mean prediction. The model is reset with the original data at
    the end of collecting a batch but if you use a model where training the model with the same data leads to different
    predictions, the model behaviour will be modified.
    """

    def __init__(
        self,
        model: IModel,
        acquisition: Acquisition,
        acquisition_optimizer: AcquisitionOptimizerBase,
        batch_size: int = 1,
    ):
        """
        :param model: Model that is used by the acquisition function
        :param acquisition: Acquisition to be optimized to find each point in batch
        :param acquisition_optimizer: Acquisition optimizer that optimizes acquisition function
                                      to find each point in batch
        :param batch_size: Number of points to calculate in batch
        """
        if (not isinstance(batch_size, int)) or (batch_size < 1):
            raise ValueError("Batch size should be a positive integer")
        self.model = model
        self.acquisition = acquisition
        self.acquisition_optimizer = acquisition_optimizer
        self.batch_size = batch_size

    def compute_next_points(self, loop_state: LoopState, context: dict = None) -> np.ndarray:
        """
        :param loop_state: Object containing history of the loop
        :param context: Contains variables to fix through optimization of acquisition function. The dictionary key is
                        the parameter name and the value is the value to fix the parameter to.
        :return: 2d array of size (batch_size x input dimensions) of new points to evaluate
        """
        self.acquisition.update_parameters()
        new_xs = []
        original_data = (self.model.X, self.model.Y)
        for _ in range(self.batch_size):
            new_x, _ = self.acquisition_optimizer.optimize(self.acquisition, context)
            new_xs.append(new_x)
            new_y = self.model.predict(new_x)[0]

            # Add new point as fake observation in model
            all_x = np.concatenate([self.model.X, new_x], axis=0)
            all_y = np.concatenate([self.model.Y, new_y], axis=0)
            self.model.set_data(all_x, all_y)

        # Reset data
        self.model.set_data(*original_data)
        return np.concatenate(new_xs, axis=0)


class RandomSampling(CandidatePointCalculator):
    """
    Samples a new candidate point uniformly at random

    """

    def __init__(self, parameter_space: ParameterSpace):
        """
        :param parameter_space: Input space
        """
        self.parameter_space = parameter_space

    def compute_next_points(self, loop_state: LoopState, context: dict = None) -> np.ndarray:
        """
        :param loop_state: Object that contains current state of the loop
        :param context: Contains variables to fix through optimization of acquisition function. The dictionary key is
                        the parameter name and the value is the value to fix the parameter to.
        :return: (1 x n_dims) array of next inputs to evaluate the function at
        """

        if context is not None:
            context_manager = ContextManager(self.parameter_space, context)
            sample = context_manager.contextfree_space.sample_uniform(1)
            sample = context_manager.expand_vector(sample)
        else:
            sample = self.parameter_space.sample_uniform(1)
        return sample



import collections
import time
from typing import Callable, Dict, List, Tuple

import GPy
import numpy as np
import numpy.linalg as la
import scipy as sp
from GPy.util.univariate_Gaussian import cdfNormal, derivLogCdfNormal, logCdfNormal, std_norm_cdf, std_norm_pdf
from scipy.stats import multivariate_normal, mvn, norm

from .. import ComparisonGP, ComparisonGPEmukitWrapper, util
from .acquisition_function import AcquisitionFunction


class SequentialThompsonSampler:
    """
    A class for performing Thompson sampling from a GP. The sampler works so that
    it stores the simulated draws and conditions the new draws on (part of) them.

    :param model: A model from which posterior the samples are drawn from
    :param seed: Random seed that specifies the random sample
    :param delta: Defines how close the samples for numerical derivatives are taken from
    :param num_points: Number of points the samples are conditioned on
    """

    def __init__(self, model: ComparisonGP, seed: float = None, delta: float = 1e-5, num_points: int = None):
        self.model = model  # model
        self.posterior = model.posterior

        self.d = self.model.X.shape[1]  # dimensionality of the input points

        self.seeds = []
        self.reset(seed=seed)

        self.delta = delta
        if num_points is None:
            num_points = int(100 * round(np.sqrt(self.d)))
        self.num_points = num_points
        self.scaling = np.array([model.kern.lengthscale[:]]).flatten()

    def reset(self, seed: float = None) -> None:
        """
        Reset the sampler by forgetting the already drawn samples and resetting the seed.

        :param seed: new seed after reset
        """
        self.x_evaluations = np.empty((0, self.model.X.shape[1]))
        self.f_evaluations = np.empty((0, 1))
        if seed is None:
            max = 1000000
            if self.model.name == "MCMC":
                max = self.model.posterior.samples["f"].shape[0]
            seed = np.random.randint(max)
            # Check if we already have used this seed
            while seed in self.seeds:
                seed = np.random.randint(max)
        self.seeds = self.seeds + [seed]
        np.random.seed(seed)

    def add_data(self, x: np.ndarray, f: np.ndarray) -> None:
        """
        Add new points and evaluations to the stack of already evaluated points

        :param x: new locations
        :param f: new evaluations
        """
        self.x_evaluations = np.concatenate((self.x_evaluations, x), axis=0)
        self.f_evaluations = np.concatenate((self.f_evaluations, f), axis=0)

    def get_mu_sigma(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the posterior predictive distribution at the given locations without conditionin on the already evaluated locatons

        :param x: New locations where the distribution is wanted at
        :return: Tuple containing the predictive mean and covariance
        """
        if self.model.name == "MCMC":
            mu, L = self.posterior._get_mu_L(x, with_index=self.seeds[-1])
            Sigma = L[0, :, :] @ L[0, :, :].T
        else:
            mu, Sigma = self.model.predict_noiseless(x, full_cov=True)
            if len(Sigma.shape) > 2:
                Sigma = Sigma[:, :, 0]
        mu = mu.reshape(-1, 1)
        Sigma = Sigma + 1e-7 * np.eye(Sigma.shape[0])  # noise for robustness
        return mu, Sigma

    def get_posterior_predictive_distribution(
        self, x: np.ndarray, inds: int = np.arange(0)
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the predictive distribution at the given locations by conditionin on old points and the posterior

        :param x: New location where the distribution is wanted at
        :param inds: indices of already evaluated points the distribution is conditioned with
        :return: Tuple containing the predictive mean and covariance
        """
        n1, n2 = inds.shape[0], x.shape[0]
        mu, Sigma = self.get_mu_sigma(np.concatenate((self.x_evaluations[inds, :].reshape(-1, self.d), x), axis=0))
        if n1 > 0:
            kXX = Sigma[np.ix_(range(n1), range(n1))]
            kxX = Sigma[np.ix_(range(n1, n1 + n2), range(n1))]
            kxx = Sigma[np.ix_(range(n1, n1 + n2), range(n1, n1 + n2))]
            kXX_inv = la.inv(kXX)
            mu_pred = mu[n1:] + kxX @ kXX_inv @ (self.f_evaluations[inds] - mu[:n1])
            Sigma_pred = kxx - kxX @ kXX_inv @ kxX.T
            return mu_pred, Sigma_pred
        else:
            return mu, Sigma

    def get_posterior_predictive_distribution_robust(self, x, num_points=100) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the predictive distribution at the given locations by conditionin on old points and the posterior

        :param x: new point the sample is wanted to evaluate at
        :param num_points: number of points taken into account when computing the predictive distribution
        :return: Tuple containing the predictive mean and covariance
        """
        N, n = self.x_evaluations.shape[0], x.shape[0]
        inds = np.arange(N)
        if N >= num_points:
            inds = np.empty((num_points, n), dtype=int)
            for i in range(n):
                dists = np.linalg.norm(
                    (
                        self.x_evaluations
                        - x[
                            None,
                            i,
                            :,
                        ]
                    )
                    / self.scaling[None, :],
                    axis=1,
                )
                inds[:, i] = np.argsort(dists)[:num_points]
            inds = np.unique(inds.flatten())
        return self.get_posterior_predictive_distribution(x, inds=inds)

    def evaluate_and_add(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the value and gradient of the sample at given point

        :param x: new point the sample is wanted to evaluate at
        :return: Tuple containing the value and gradient
        """
        X = np.atleast_2d(X).reshape((-1, self.d))
        ff, fgrad = np.empty((0)), np.empty((0, self.d))
        for i in range(X.shape[0]):
            # deviate to compute the numerical gradients
            x = np.tile(X[i, :].reshape((1, self.d)), (self.d + 1, 1))
            for i in range(self.d):
                x[i + 1, i] = x[i + 1, i] + self.delta
            # Get predictive distribution
            mu, k = self.get_posterior_predictive_distribution_robust(x, num_points=self.num_points)

            # evaluate and add
            f = np.random.multivariate_normal(mu.flatten(), k).reshape((-1, 1))
            self.add_data(x, f)

            # Evaluate numerical gradients
            grad = np.empty((1, self.d))
            for i in range(self.d):
                grad[0, i] = (f[i + 1] - f[0]) / self.delta
            ff = np.concatenate((ff, f[0]), axis=0)
            fgrad = np.concatenate((fgrad, grad), axis=0)
        return ff.reshape((-1, 1)), fgrad


class ThompsonSampling(AcquisitionFunction):
    def acq_fun_optimizer(
        self, m: ComparisonGP, bounds: np.ndarray, batch_size: int, get_logger: Callable
    ) -> np.ndarray:
        """
        Implements the optimization scheme for the Thompson sampling acquisition function

        :param m: The model which posterior is used by the acquisition function (from which the samples are drawn from)
        :param bounds: the optimization bounds of the new sample
        :param batch_size: How many points are there in the batch
        :param get_logger: Function for receiving the legger where the prints are forwarded.
        :return: optimized locations
        """
        self.sampler = SequentialThompsonSampler(m)
        X_all = np.empty((batch_size, m.X.shape[1]))
        for i in range(batch_size):
            self.sampler.reset()  # reset the sampling seed
            yb = np.inf
            Xb = None
            for j in range(self.acq_opt_restarts):
                try:
                    # Initial point of the optimization
                    X0 = util.random_sample(bounds, 1)
                    res = sp.optimize.minimize(
                        fun=self.sampler.evaluate_and_add,
                        x0=X0.flatten(),
                        method="L-BFGS-B",
                        jac=True,
                        bounds=bounds,
                        options=self.optimizer_options,
                    )
                    X = res.x.reshape(1, len(bounds))
                    y = res.fun
                    if y < yb:
                        yb = y
                        Xb = X
                except Exception as e:
                    get_logger().error("Solver failed. Below the exception\n{}".format(e))
            assert Xb is not None
            X_all[i, :] = Xb
        return X_all

    def reset(self, model: ComparisonGPEmukitWrapper) -> None:
        """
        Some acquisition functions need to be reseted, this method is for that.
        :param m: the GP model which posterior is used
        """
        try:
            self.sampler.reset()
        except:
            self.sampler = SequentialThompsonSampler(model)

    def evaluate(self, x: np.ndarray, model: ComparisonGPEmukitWrapper) -> np.ndarray:
        """
        Computes the Expected Improvement.
        :param x: points where the acquisition is evaluated.
        :param m: the GP model which posterior is used
        :return: acquisition function value
        """
        return self.sampler.evaluate_and_add(x)[0]

    def evaluate_with_gradients(self, x: np.ndarray, model: ComparisonGPEmukitWrapper) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the Expected Improvement.
        :param x: points where the acquisition is evaluated.
        :param model: the GP model which posterior is used
        :return: A tuple containing the acquisition function values and their gradients
        """
        return self.sampler.evaluate_and_add(x)



"""The WSABI-L loop"""


from ...core.loop import FixedIntervalUpdater, ModelUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import create_loop_state
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace
from ..acquisitions import UncertaintySampling
from ..methods import WSABIL


class WSABILLoop(OuterLoop):
    """The loop for WSABI-L.

    .. rubric:: References

    Gunter et al. 2014 *Sampling for Inference in Probabilistic Models with Fast Bayesian Quadrature*,
    Advances in Neural Information Processing Systems (NeurIPS), 27, pp. 2789–2797.

    .. seealso::
        :class:`emukit.quadrature.methods.WSABIL`

    :param model: The WSABI-L model.
    :param model_updater: Defines how and when the quadrature model is updated if new data arrives.
                          Defaults to updating hyper-parameters every iteration.
    :param acquisition_optimizer: Optimizer selecting next evaluation points by maximizing acquisition.
                                  Gradient based optimizer is used if None. Defaults to None.
    """

    def __init__(
        self, model: WSABIL, model_updater: ModelUpdater = None, acquisition_optimizer: AcquisitionOptimizerBase = None
    ):

        # WSABI-L is used with uncertainty sampling.
        acquisition = UncertaintySampling(model, measure_power=1)

        if model_updater is None:
            model_updater = FixedIntervalUpdater(model, 1)

        space = ParameterSpace(model.reasonable_box_bounds.convert_to_list_of_continuous_parameters())
        if acquisition_optimizer is None:
            acquisition_optimizer = GradientAcquisitionOptimizer(space)
        candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)
        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updater, loop_state)

        self.model = model



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


"""
Contains linear models
"""

import GPy
import numpy as np


class GPyLinearMultiFidelityModel(GPy.core.GP):
    """
    A thin wrapper around GPy.core.GP that does some input checking and provides a default likelihood
    """

    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        kernel: GPy.kern.Kern,
        n_fidelities: int,
        likelihood: GPy.likelihoods.Likelihood = None,
    ):
        """

        :param X: Training data features with fidelity input appended as last column
        :param Y: Training data targets
        :param kernel: Multi-fidelity kernel
        :param n_fidelities: Number of fidelities in problem
        :param likelihood: GPy likelihood object.
                           Defaults to MixedNoise which has different noise levels for each fidelity
        """

        # Input checks
        if not isinstance(X, np.ndarray):
            raise ValueError("X should be an array")

        if not isinstance(Y, np.ndarray):
            raise ValueError("Y should be an array")

        if X.ndim != 2:
            raise ValueError("X should be 2d")

        if Y.ndim != 2:
            raise ValueError("Y should be 2d")

        if np.any(X[:, -1] >= n_fidelities):
            raise ValueError("One or more points has a higher fidelity index than number of fidelities")

        # Make default likelihood as different noise for each fidelity
        if likelihood is None:
            likelihood = GPy.likelihoods.mixed_noise.MixedNoise(
                [GPy.likelihoods.Gaussian(variance=1.0) for _ in range(n_fidelities)]
            )
        y_metadata = {"output_index": X[:, -1].astype(int)}
        super().__init__(X, Y, kernel, likelihood, Y_metadata=y_metadata)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Union

import numpy as np

from ...core.interfaces.models import IDifferentiable, IModel
from ...quadrature.interfaces.base_gp import IBaseGaussianProcess
from ..measures import BoxDomain, IntegrationMeasure
from .warpings import Warping


class WarpedBayesianQuadratureModel(IModel, IDifferentiable):
    r"""The general class for Bayesian quadrature (BQ) with a warped Gaussian process model.

    The model is of the form :math:`f = w(g(x))` where :math:`g\sim\mathcal{GP}` is a Gaussian process
    and :math:`w:\mathbb{R}\rightarrow\mathbb{R}` is a deterministic warping function.

    Inherit from this class to create new warped Bayesian quadrature models.

    .. seealso::
        * :class:`emukit.quadrature.methods.warpings.Warping`
        * :class:`emukit.quadrature.methods.VanillaBayesianQuadrature`
        * :class:`emukit.quadrature.methods.BoundedBayesianQuadrature`
        * :class:`emukit.quadrature.methods.WSABIL`

    """

    def __init__(self, base_gp: IBaseGaussianProcess, warping: Warping, X: np.ndarray, Y: np.ndarray):
        """
        :param base_gp: The underlying Gaussian process model.
        :param warping: The warping of the underlying Gaussian process model.
        :param X: The initial locations of integrand evaluations, shape (n_points, input_dim).
        :param Y: The values of the integrand at X, shape (n_points, 1).
        """
        self._warping = warping
        self.base_gp = base_gp
        # set data to ensure that the base_gp get the correctly transformed observations.
        self.set_data(X, Y)

    @property
    def X(self) -> np.ndarray:
        """The data nodes."""
        return self.base_gp.X

    @property
    def Y(self) -> np.ndarray:
        """The data evaluations at the nodes."""
        return self._warping.transform(self.base_gp.Y)

    @property
    def integral_bounds(self) -> Union[None, BoxDomain]:
        """The integration bounds. ``None`` if integration domain is not bounded."""
        return self.base_gp.kern.integral_bounds

    @property
    def reasonable_box_bounds(self) -> BoxDomain:
        """Reasonable box bounds.

        This box is used by the acquisition optimizer even when ``integral_bounds`` is ``None``.
        By default it is set to :meth:`get_box()` of the integration measure used, or, if not available,
        to the ``integral_bounds``.

        .. seealso::
            :class:`emukit.quadrature.measures.IntegrationMeasure.get_box`

        """
        return self.base_gp.kern.reasonable_box

    @property
    def measure(self) -> Union[None, IntegrationMeasure]:
        """The measure used for integration. ``None`` for standard Lebesgue measure."""
        return self.base_gp.kern.measure

    def transform(self, Y: np.ndarray) -> np.ndarray:
        """The transform from base-GP to integrand implicitly defined by the warping used."""
        return self._warping.transform(Y)

    def inverse_transform(self, Y: np.ndarray) -> np.ndarray:
        """The transform from integrand to base-GP implicitly defined by the warping used."""
        return self._warping.inverse_transform(Y)

    def predict_base(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Compute predictive means and variances of the warped GP as well as the base GP.

        :param X_pred: Locations at which to predict, shape (n_points, input_dim).
        :returns: Predictive mean and variances of warped GP, and predictive mean and variances of base-GP in that order
                  all shapes (n_points, 1).
        """
        raise NotImplementedError

    def predict_base_with_full_covariance(
        self, X_pred: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Compute predictive means and covariance of the warped GP as well as the base GP.

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :returns: Predictive mean and covariance of warped GP, predictive mean and covariance of base-GP in that order.
                  mean shapes both (n_points, 1) and covariance shapes both (n_points, n_points)
        """
        raise NotImplementedError

    def predict_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute predictive means and covariance of warped GP.

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :return: predictive mean, predictive full covariance of warped-GP, shapes (n_points, 1) and (n_points, n_points)
        """
        return self.predict_base_with_full_covariance(X_pred)[:2]

    def predict(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute predictive means and variances of warped-GP.

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :return: predictive mean, predictive variances of warped-GP, both shapes (n_points, 1)
        """
        return self.predict_base(X_pred)[:2]

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """Set the new data in the model.

        First, potential warping parameters that are not being optimized but do depend on the data
        in an analytic way are updated. This is done via the method :meth:`compute_warping_params`.
        Then, the new data is automatically transformed and set in the model.

        :param X: Observation locations, shape (n_points, input_dim).
        :param Y: Integrand observations at X, shape (n_points, 1).
        """
        self._warping.update_parameters(**self.compute_warping_params(X, Y))
        self.base_gp.set_data(X, self._warping.inverse_transform(Y))

    def compute_warping_params(self, X: np.ndarray, Y: np.ndarray) -> dict:
        """Compute new parameters of the warping that are dependent on data, and that are not being optimized.

        This method is called by default when new data is being set in :meth:`set_data`.
        By default, this method returns an empty dict (no warping params need to be updated).
        Override this method in case warping parameters are data dependent.

        .. seealso::
            :class:`emukit.quadrature.methods.warpings.Warping.update_parameters`

        :param X: Observation locations, shape (n_points, input_dim).
        :param Y: Integrand observations at X, shape (n_points, 1).
        :return: Dictionary containing new warping parameters. Names of parameters are the keys.
        """
        return {}

    def optimize(self) -> None:
        """Optimizes the hyperparameters of the base GP."""
        self.base_gp.optimize()

    def integrate(self) -> Tuple[float, float]:
        """Compute an estimator of the integral as well as its variance.

        :returns: Estimator of integral and its variance.
        """
        raise NotImplementedError

    @staticmethod
    def symmetrize_matrix(A: np.ndarray) -> np.ndarray:
        r"""Symmetrize a matrix.

        The symmetrized matrix is computed as :math:`A_{sym} = \frac{1}{2} (A + A^{\intercal})`.

        :param A: The square matrix :math:`A`, shape (N, N)
        :return: The symmetrized matrix :math:`A_{sym}`, shape (N, N).
        """
        return 0.5 * (A + A.T)



from typing import Union

import numpy as np

from emukit.bayesian_optimization.acquisitions import NegativeLowerConfidenceBound
from emukit.core.interfaces import IDifferentiable, IModel


class DynamicNegativeLowerConfidenceBound(NegativeLowerConfidenceBound):
    def __init__(self, model: Union[IModel, IDifferentiable], input_space_size: int, delta: float) -> None:
        """
        Dynamic extension of the LCB acquisition. The beta coefficient is updated at each iteration, based on the explorativeness parameter delta which is inversely
        proportional to beta itself - the higher the delta the less explorative the selection will be.
        Please consider that regret bounds based on the dynamic exploration coefficient only hold for selected kernel classes exhibiting boundedness and smoothness.
        See the base class for paper references.
        This class may also be taken as a reference for acquisition functions that dynamically update their parameters thanks to the update_parameters() hook; the implicit assumption is that this method is invoked once per iteration (it is no big deal if this is the case for a constant number of times per iteration; should it be more then we are increasing the beta too fast).
        :param model: The underlying model that provides the predictive mean and variance for the given test points
        :param input_space_size: the size of the finite D grid on which the function is evaluated
        :param delta: the exploration parameter determining the beta exploration coefficient; delta must be in (0, 1) and it is inversely related to beta
        """
        assert input_space_size > 0, "Invalid dimension provided"
        assert 0 < delta < 1, "Delta must be in (0, 1)"
        super().__init__(model)
        self.input_space_size = input_space_size
        self.delta = delta
        self.iteration = 0

    def optimal_beta_selection(self) -> float:
        return 2 * np.log(self.input_space_size * (self.iteration**2) * (np.pi**2) / (6 * self.delta))

    def update_parameters(self) -> None:
        self.iteration += 1
        self.beta = self.optimal_beta_selection()



import itertools
import os
import traceback
from copy import deepcopy
from typing import Callable, Dict, List, Tuple

import GPy
import numpy as np
import paramz
import stan_utility
from GPy.core.parameterization.param import Param
from GPy.likelihoods import Gaussian
from GPy.util import choleskies
from GPy.util.linalg import dpotrs, dtrtrs, jitchol, pdinv, tdot

import emukit
from emukit.core.interfaces import IModel

from . import util
from .inferences import StanPosterior
from .inferences import ep_batch_comparison as ep
from .inferences import vi_batch_comparison as vi


class ComparisonGP(GPy.core.Model):
    """
    A class for all common methods needed for the different ComparisonGP wrappers
    """

    def get_current_best(self) -> float:
        """
        :return: minimum of means of predictions at all input locations (needed by q-EI)
        """
        return min(self.Y)

    def get_y_pred(self) -> np.ndarray:
        """
        :return: GP mean at inputs used to compute the posterior approximation (needed by q-EI)
        """
        y_pred, _ = self.predict(self.X, include_likelihood=False)
        return y_pred

    def log_likelihood(self) -> float:
        """
        :return: log marginal likelihood needed for optimizing hyper parameters and performing model comparison
        """
        return self._log_marginal_likelihood

    def predict(
        self, Xnew: np.ndarray, full_cov: bool = False, include_likelihood=True
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predictive mean and covariance of the GP at the input location
        :param Xnew: Locations the prections are wanted at
        :param full_cov: If the user wants the function to return the full covariance or only the diagonal
        :param include_likelihood: If the user wants the function to add the noise of the observations to the prediction covariance
        :return: predictive mean and predictive covariance
        """
        pred_mean, pred_var = self.posterior._raw_predict(
            self.kern, Xnew, self.X, full_cov=full_cov
        )  # self.posterior._raw_predict(self.kern, np.hstack([Xnew,ki]), np.hstack([self.X, self.ki]), full_cov=full_cov)
        if include_likelihood:
            pred_var = pred_var + self.likelihood.variance
        return pred_mean, pred_var

    def predict_noiseless(self, Xnew, full_cov=False) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predictive mean and covariance of the GP latent function at the input location
        :param Xnew: Locations the prections are wanted at
        :param full_cov: If the user wants the function to return the full covariance or only the diagonal
        :return: predictive latent mean and predictive latent covariance
        """
        return self.predict(Xnew, full_cov=full_cov, include_likelihood=False)

    def posterior_samples_f(self, X: np.ndarray, size: int = 10, **predict_kwargs) -> np.ndarray:
        """
        Draw random samples from the posterior predictive distribution

        :param X: Locations where the posterior samples should be drawn at
        :param size: Number of posterior samples
        :return: Simulated posterior samples
        """
        predict_kwargs["full_cov"] = True  # Always use the full covariance for posterior samples.
        predict_kwargs["include_likelihood"] = False
        m, v = self.predict(X, **predict_kwargs)

        def sim_one_dim(m, v):
            # Draw posterior sample in one dimension
            return np.random.multivariate_normal(m, v, size).T

        if self.output_dim == 1:
            return sim_one_dim(m.flatten(), v)[:, np.newaxis, :]
        else:
            fsim = np.empty((X.shape[0], self.output_dim, size))
            for d in range(self.output_dim):
                if v.ndim == 3:
                    fsim[:, d, :] = sim_one_dim(m[:, d], v[:, :, d])
                else:
                    fsim[:, d, :] = sim_one_dim(m[:, d], v)
        return fsim


class EPComparisonGP(ComparisonGP):
    """
    GPy wrapper for a GP model consisting of preferential batch observations when the posterior is approximated using Expectation Propagation

    :param X: All locations of both direct observations and batch comparisons
    :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param kernel: A GPy kernel used
    :param likelihood: A GPy likelihood. Only Gaussian likelihoods are accepted
    :param name: Name of the model. Defaults to 'EpComparisonGP'
    :param ep_max_it: Maximum number of iterations used when approximating the posterior in EP.
    :param eta: parameter for fractional EP updates.
    :param delta: damping EP updates factor.
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    """

    def __init__(
        self,
        X: np.ndarray,
        y: List[Tuple[int, float]],
        yc: List[List[Tuple[int, int]]],
        kernel: GPy.kern.Kern,
        likelihood: GPy.likelihoods.Gaussian,
        name: str = "EPComparisonGP",
        ep_max_itt: int = 100,
        delta: float = 0.5,
        eta: float = 0.5,
        get_logger: Callable = None,
    ):

        super(EPComparisonGP, self).__init__(name=name)

        self.N, self.D = X.shape[0], X.shape[1]

        self.output_dim = 1  # hard coded, the code doesn't support multi output case

        self.X = X
        self.y = y
        self.yc = yc

        self.kern = kernel
        self.likelihood = likelihood

        # A helper parameter for EP. Each observation could possibly come from different kernels and likelihoods.
        # The inference supports this already, but this GPy wrapper doesn't
        self.sigma2s = self.likelihood.variance * np.ones((X.shape[0], 1), dtype=int)

        self.ep_max_itt = ep_max_itt
        self.eta = eta
        self.delta = delta

        self.link_parameter(self.kern)
        self.link_parameter(self.likelihood)
        self.posterior, self.ga_approx, self.Y = None, None, None
        self.get_logger = get_logger

    def parameters_changed(self):
        """
        Update the posterior approximation after kernel or likelihood parameters have changed or there are new observations
        """
        # Recompute the posterior approximation
        self.posterior, self._log_marginal_likelihood, self.grad_dict, self.ga_approx = ep.ep_comparison(
            self.X,
            self.y,
            self.yc,
            self.kern,
            self.sigma2s,
            max_itt=self.ep_max_itt,
            tol=1e-6,
            delta=self.delta,
            eta=self.eta,
            ga_approx_old=self.ga_approx,
            get_logger=self.get_logger,
        )

        # predict Y at inputs (needed by q-EI)
        self.Y = self.get_y_pred()

    def set_XY(self, X: np.ndarray, y: List[Tuple[int, float]], yc: List[List[Tuple[int, int]]]):
        """
        Set new observations and recompute the posterior

        :param X: All locations of both direct observations and batch comparisons
        :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
        :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
        """
        self.X = X
        self.y = y
        self.yc = yc
        self.sigma2s = self.likelihood.variance * np.ones((X.shape[0], 1), dtype=int)
        self.parameters_changed()


class VIComparisonGP(ComparisonGP):
    """
    GPy wrapper for a GP model consisting of preferential batch observations when the posterior is approximated using Variational Inference

    :param X: All locations of both direct observations and batch comparisons
    :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param kernel: A GPy kernel used
    :param likelihood: A GPy likelihood. Only Gaussian likelihoods are accepted
    :param vi_mode: A string indicating if to use full rank or mean field VI
    :param name: Name of the model. Defaults to 'VIComparisonGP'
    :param max_iters: Maximum number of iterations used when approximating the posterior in VI.
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    """

    def __init__(
        self,
        X: np.ndarray,
        y: List[Tuple[int, float]],
        yc: List[List[Tuple[int, int]]],
        kernel: GPy.kern.Kern,
        likelihood: Gaussian,
        vi_mode: str = "fr",
        name: str = "VIComparisonGP",
        max_iters: int = 50,
        get_logger: Callable = None,
    ):
        super(VIComparisonGP, self).__init__(name=name)

        self.N, self.D = X.shape[0], X.shape[1]

        self.output_dim = 1
        self.get_logger = get_logger
        self.X = X
        self.y = y
        self.yc = yc

        self.max_iters = max_iters
        self.vi_mode = vi_mode

        self.kern = kernel
        self.likelihood = likelihood

        self.sigma2s = self.likelihood.variance * np.ones((X.shape[0], 1), dtype=int)
        jitter = 1e-6
        K = self.kern.K(X)
        L = np.linalg.cholesky(K + np.identity(K.shape[0]) * jitter)

        self.alpha = np.zeros((self.N, 1))
        self.beta = np.ones((self.N, 1))

        self.posterior = None

        # If we are using full rank VI, we initialize it with mean field VI
        if self.vi_mode == "FRVI":
            self.posterior, _, _, self.alpha, self.beta = vi.vi_comparison(
                self.X, self.y, self.yc, self.kern, self.sigma2s, self.alpha, self.beta, max_iters=50, method="mf"
            )
            self.beta = choleskies._triang_to_flat_pure(jitchol(self.posterior.covariance)[None, :])

    def parameters_changed(self):
        """
        Update the posterior approximation after kernel or likelihood parameters have changed or there are new observations
        """
        if self.vi_mode == "fr":
            method = "fr"
        else:
            method = "mf"

        self.posterior, self._log_marginal_likelihood, self.grad_dict, alpha, beta = vi.vi_comparison(
            self.X,
            self.y,
            self.yc,
            self.kern,
            self.sigma2s,
            self.alpha,
            self.beta,
            max_iters=self.max_iters,
            method=method,
            get_logger=self.get_logger,
        )

        self.alpha = alpha
        self.beta = beta
        self.Y = self.get_y_pred()

    def set_XY(self, X: np.ndarray, y: List[Tuple[int, float]], yc: List[List[Tuple[int, int]]]):
        """
        Set new observations and recompute the posterior

        :param X: All locations of both direct observations and batch comparisons
        :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
        :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
        """
        self.N, self.D = X.shape[0], X.shape[1]
        self.X = X
        self.y = y
        self.yc = yc
        self.sigma2s = self.likelihood.variance * np.ones((X.shape[0], 1), dtype=int)
        alpha = np.zeros((self.N, 1))
        if self.vi_mode == "fr":
            beta = np.ones((int((self.N**2 + self.N) / 2), 1))
        else:
            beta = np.ones((self.N, 1))
        for i in range(len(self.alpha)):
            alpha[i] = self.alpha[i]
        for i in range(len(self.beta)):
            beta[i] = self.beta[i]
        self.alpha, self.beta = alpha, beta
        self.posterior = None
        self.parameters_changed()


class VIComparisonGPMF(VIComparisonGP):
    """
    A very thin wrapper for mean field VI
    """

    def __init__(
        self,
        X: np.ndarray,
        y: List[Tuple[int, float]],
        yc: List[List[Tuple[int, int]]],
        kernel: GPy.kern.Kern,
        likelihood: Gaussian,
        name: str = "VIComparisonGPMF",
        max_iters: int = 50,
        get_logger: Callable = None,
    ):
        super(VIComparisonGPMF, self).__init__(
            X, y, yc, kernel, likelihood, name=name, max_iters=max_iters, vi_mode="mf", get_logger=get_logger
        )


class VIComparisonGPFR(VIComparisonGP):
    """
    A very thin wrapper for full rank VI
    """

    def __init__(
        self,
        X: np.ndarray,
        y: List[Tuple[int, float]],
        yc: List[List[Tuple[int, int]]],
        kernel: GPy.kern.Kern,
        likelihood: Gaussian,
        name: str = "VIComparisonGPFR",
        max_iters: int = 50,
        get_logger: Callable = None,
    ):
        super(VIComparisonGPFR, self).__init__(
            X, y, yc, kernel, likelihood, name=name, max_iters=max_iters, vi_mode="fr", get_logger=get_logger
        )


class MCMCComparisonGP(ComparisonGP):
    """
    GPy wrapper for a GP model consisting of preferential batch observations when the posterior is approximated using MCMC smaples

    :param X: All locations of both direct observations and batch comparisons
    :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
    :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
    :param kernel: A GPy kernel used
    :param likelihood: A GPy likelihood. Only Gaussian likelihoods are accepted
    :param posterior_samples: Number of posterior samples used for approximation
    :param vi_mode: A string indicating if to use full rank or mean field VI
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    """

    def __init__(
        self,
        X: np.ndarray,
        y: List[Tuple[int, float]],
        yc: List[List[Tuple[int, int]]],
        kernel: GPy.kern.Kern,
        likelihood: Gaussian,
        posterior_samples: int = 45,
        get_logger: Callable = None,
    ):
        super(MCMCComparisonGP, self).__init__(name="MCMC")
        self.N, self.D = X.shape[0], X.shape[1]
        self.output_dim = 1  ## hard coded
        self.posterior_samples = posterior_samples
        self.X = X
        self.y = y
        self.yc = yc
        self.noise_std = np.sqrt(likelihood.variance[0])
        self.sigma2s = (likelihood.variance[0]) * np.ones((X.shape[0], 1), dtype=int)
        self.variance = kernel.variance[0]
        self.lengthscale = np.array([kernel.lengthscale[:]]).flatten()
        self.kern = kernel
        self.posterior = None
        if not os.path.exists(stan_utility.file_utils.get_path_of_cache()):
            os.makedirs(stan_utility.file_utils.get_path_of_cache())
        self.model = stan_utility.compile_model(
            os.path.join(os.path.dirname(__file__), "inferences/sexpgp_comparison.stan"), model_name="comparison_model"
        )
        self.get_logger = get_logger
        self.parameters_changed()

    def _fit_stan(self) -> Dict:
        """
        Fit stan using the given observations and return a wrapper object that allows a simple framework for handling the posterior samples

        :return: A pystan fit object containing the posterior fit for the model
        """

        def inits():
            return {"eta": np.zeros((self.X.shape[0],))}

        dat = {
            "N": len(self.y),
            "N_comp": len(list(itertools.chain(*self.yc))),
            "Nf": self.X.shape[0],
            "d": self.X.shape[1],
            "y": np.array([yi[1] for yi in self.y]).flatten(),
            "yi": np.array([yi[0] + 1 for yi in self.y], dtype=int),
            "y_comp1": np.array([i[0] + 1 for i in list(itertools.chain(*self.yc))], dtype=int),
            "y_comp2": np.array([i[1] + 1 for i in list(itertools.chain(*self.yc))], dtype=int),
            "x": self.X,
            "rho": self.lengthscale,
            "alpha": np.sqrt(self.variance),
            "sigma": self.noise_std,
            "delta": 1e-9,
        }
        iter = 3000
        chains = 6
        warmup = iter - int(self.posterior_samples // chains)
        if self.get_logger is not None:
            self.get_logger().info("Starting to sample from the posterior")
        fit = self.model.sampling(
            data=dat,
            seed=194838,
            chains=6,
            n_jobs=1,
            iter=iter,
            warmup=warmup,
            refresh=-1,
            control={"adapt_delta": 0.9999, "max_treedepth": 12},
            init=inits,
        )
        if self.get_logger is not None:
            self.get_logger().info("Sampling finished")
        return fit

    def parameters_changed(self):
        """
        Update the posterior approximation after kernel or likelihood parameters have changed or there are new observations
        """
        # Run Stan:
        fit = self._fit_stan()
        self.samples = fit.extract(permuted=True)
        remove_levels = True
        if len(self.y) > 0:
            remove_levels = False
        self.posterior = StanPosterior(
            self.samples,
            self.X,
            self.kern,
            self.noise_std,
            Y=np.array([yi[1] for yi in self.y]).flatten(),
            remove_levels=remove_levels,
            y=self.y,
            yc=self.yc,
            get_logger=self.get_logger,
        )
        self._log_marginal_likelihood, self.grad_dict = None, {}
        self.y_pred = self.get_y_pred()
        self.Y = self.y_pred

    def set_XY(self, X: np.ndarray, y: List[Tuple[int, float]], yc: List[List[Tuple[int, int]]]):
        """
        Set new observations and recompute the posterior

        :param X: All locations of both direct observations and batch comparisons
        :param y: Direct observations in as a list of tuples telling location index (row in X) and observation value.
        :param yc: Batch comparisons in a list of lists of tuples. Each batch is a list and tuples tell the comparisons (winner index, loser index)
        """
        self.X = X
        self.y = y
        self.yc = yc
        self.parameters_changed()


class ComparisonGPEmukitWrapper(IModel):
    """
    This is a thin wrapper around ComparisonGP to allow users to plug GPy models into Emukit
    """

    def __init__(self, gpy_model: ComparisonGP, batch_size=1):
        """
        :param gpy_model: GPy model object to wrap
        :param batch_size: Number of samples in a batch
        """
        self.model = gpy_model
        self.batch_size = batch_size
        self.g = 0

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 of the predictive distribution at each input location
        """
        return self.model.predict(X)

    def predict_with_full_covariance(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 and n_points x n_points of the predictive
                 mean and variance at each input location
        """
        return self.model.predict(X, full_cov=True)

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model

        :param X: New training features
        :param Y: New training outputs
        """
        if X.shape[0] > 0:
            X = X[0, :].reshape((1, -1))
            Y = Y[0, :].reshape((-1))
            ind = self.model.X.shape[0]
            X_set = self.model.X
            y_set = self.model.y
            yc_set = self.model.yc
            if X.shape[1] == self.model.X.shape[1]:  # Direct observation
                X_set = np.concatenate((self.model.X, X), axis=0)
                y_set += [(i + ind, Y[i]) for i in range(len(Y))]
            else:
                X_tmp = np.concatenate(
                    tuple(X[:, i * self.batch_size : (i + 1) * self.batch_size] for i in range(len(Y))), axis=0
                )
                X_set = np.concatenate((self.model.X, X_tmp), axis=0)
                yc_set += [[(yc[0] + ind, yc[1] + ind) for yc in util.comparison_form(Y)]]

            self.model.set_XY(X_set, y_set, yc_set)
        #
        # self.model.set_XY(X, Y)

    def predict_covariance(self, X: np.ndarray, with_noise: bool = True) -> np.ndarray:
        """
        Calculates posterior covariance between points in X
        :param X: Array of size n_points x n_dimensions containing input locations to compute posterior covariance at
        :param with_noise: Whether to include likelihood noise in the covariance matrix
        :return: Posterior covariance matrix of size n_points x n_points
        """
        _, v = self.model.predict(X, full_cov=True, include_likelihood=with_noise)
        v = np.clip(v, 1e-10, np.inf)

        return v

    @property
    def X(self) -> np.ndarray:
        """
        :return: An array of shape n_points x n_dimensions containing training inputs
        """
        return np.empty((0, self.model.X.shape[1]))

    @property
    def Y(self) -> np.ndarray:
        """
        :return: An array of shape n_points x 1 containing training outputs
        """
        return np.empty((0, 1))

    def optimize(self):
        """
        Optimizes the model. In this case does nothing since the parameters are fixed
        """
        pass



"""The Gaussian measure."""

from typing import Union

import numpy as np

from ...core.optimization.context_manager import ContextManager
from ..typing import BoundsType
from .integration_measure import IntegrationMeasure


class GaussianMeasure(IntegrationMeasure):
    r"""The Gaussian measure.

    The Gaussian measure has density

    .. math::
        p(x)=(2\pi)^{-\frac{d}{2}} \left(\prod_{j=1}^d \sigma_j^2\right)^{-\frac{1}{2}} e^{-\frac{1}{2}\sum_{i=1}^d\frac{(x_i-\mu_i)^2}{\sigma_i^2}}

    where :math:`\mu_i` is the :math:`i` th element of the ``mean`` parameter and
    :math:`\sigma_i^2` is :math:`i` th element of the ``variance`` parameter.

    :param mean: The mean of the Gaussian measure, shape (input_dim, ).
    :param variance: The variances of the Gaussian measure, shape (input_dim, ).
                     If a scalar value is given, all dimensions will have same variance.

    :raises TypeError: If ``mean`` is not of type :class:`ndarray`.
    :raises ValueError: If ``mean`` is not of dimension 1.
    :raises TypeError: If ``variance`` is neither of type :class:`float` nor of type :class:`ndarray`.
    :raises ValueError: If ``variance`` is of type :class:`float` but is non-positive.
    :raises ValueError: If ``variance`` is of type :class:`ndarray` but of other size than ``mean``.
    :raises ValueError: If ``variance`` is of type :class:`ndarray` and any of its elements is non-positive.

    """

    def __init__(self, mean: np.ndarray, variance: Union[float, np.ndarray]):
        super().__init__(domain=None, name="GaussianMeasure")
        # check mean
        if not isinstance(mean, np.ndarray):
            raise TypeError("Mean must be of type numpy.ndarray, {} given.".format(type(mean)))

        if mean.ndim != 1:
            raise ValueError("Dimension of mean must be 1, dimension {} given.".format(mean.ndim))

        # check covariance
        is_isotropic = False

        if isinstance(variance, float):
            if variance <= 0:
                raise ValueError("Variance must be positive, current value is {}.".format(variance))
            variance = np.full((mean.shape[0],), variance)
            is_isotropic = True

        elif isinstance(variance, np.ndarray):
            if variance.shape != mean.shape:
                raise ValueError(
                    "Variance has wrong shape; {} given but {} expected.".format(variance.shape, mean.shape)
                )
            if any(variance <= 0):
                raise ValueError(
                    "All elements of variance must be positive. At least one value seems to be non positive."
                )

        else:
            raise TypeError("Variance must be of type float or numpy.ndarray, {} given.".format(type(variance)))

        self.mean = mean
        self.variance = variance
        self._input_dim = mean.shape[0]
        self.is_isotropic = is_isotropic

    @property
    def input_dim(self):
        return self._input_dim

    @property
    def full_covariance_matrix(self):
        """The full covariance matrix of the Gaussian measure."""
        return np.diag(self.variance)

    @property
    def can_sample(self) -> bool:
        return True

    def compute_density(self, x: np.ndarray) -> np.ndarray:
        factor = (2 * np.pi) ** (self.input_dim / 2) * np.prod(np.sqrt(self.variance))
        scaled_diff = (x - self.mean) / (np.sqrt(2 * self.variance))
        return np.exp(-np.sum(scaled_diff**2, axis=1)) / factor

    def compute_density_gradient(self, x: np.ndarray) -> np.ndarray:
        values = self.compute_density(x)
        diff = (x - self.mean) / self.variance
        return -diff * values[:, None]

    def reasonable_box(self) -> BoundsType:
        # The reasonable box is defined as the hypercube centered at the mean of the Gaussian with 10 standard
        # deviations expanding to either side (edge length of the cube are thus 20 standard deviations).
        # The factor 10 is somewhat arbitrary but well motivated as the Gaussian measure if virtually zero
        # outside of 10 standard deviations. See also the docstring of IntegrationMeasure.reasonable_box.
        factor = 10
        lower = self.mean - factor * np.sqrt(self.variance)
        upper = self.mean + factor * np.sqrt(self.variance)
        return list(zip(lower, upper))

    def sample(self, num_samples: int, context_manager: ContextManager = None) -> np.ndarray:
        samples = self.mean + np.sqrt(self.variance) * np.random.randn(num_samples, self.input_dim)

        if context_manager is not None:
            # Since the Gaussian is diagonal, fixing the value after sampling the joint is equal to sampling the
            # conditional.
            samples[:, context_manager.context_idxs] = context_manager.context_values

        return samples



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Union

import numpy as np

from ...core.acquisition import Acquisition
from ...core.interfaces import IDifferentiable, IModel


class NegativeLowerConfidenceBound(Acquisition):
    def __init__(self, model: Union[IModel, IDifferentiable], beta: float = 1.0) -> None:

        """
        This acquisition computes the negative lower confidence bound for a given input point. This is the same
        as optimizing the upper confidence bound if we would maximize instead of minimizing the objective function.
        For information as well as some theoretical insights see:

        Gaussian Process Optimization in the Bandit Setting: No Regret and Experimental Design
        Niranjan Srinivas, Andreas Krause, Sham Kakade, Matthias Seeger
        In Proceedings of the 27th International Conference  on  Machine  Learning

        :param model: The underlying model that provides the predictive mean and variance for the given test points
        :param beta: Is multiplied on the standard deviation to control exploration / exploitation
        """
        self.model = model
        self.beta = beta

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the negative lower confidence bound

        :param x: points where the acquisition is evaluated.
        """
        mean, variance = self.model.predict(x)
        standard_deviation = np.sqrt(variance)

        return -(mean - self.beta * standard_deviation)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple:
        """
        Computes the negative lower confidence bound and its derivative

        :param x: points where the acquisition is evaluated.
        """
        mean, variance = self.model.predict(x)
        standard_deviation = np.sqrt(variance)

        dmean_dx, dvariance_dx = self.model.get_prediction_gradients(x)
        dstandard_deviation_dx = dvariance_dx / (2 * standard_deviation)

        lcb = -(mean - self.beta * standard_deviation)

        dlcb_dx = -(dmean_dx - self.beta * dstandard_deviation_dx)

        return lcb, dlcb_dx

    @property
    def has_gradients(self):
        return isinstance(self.model, IDifferentiable)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np

from .continuous_parameter import ContinuousParameter
from .encodings import Encoding
from .parameter import Parameter


class CategoricalParameter(Parameter):
    def __init__(self, name: str, encoding: Encoding):
        super().__init__(name)

        # ensure float just in case we were given integers
        self.encoding = encoding

        self._cont_params = []
        for column_idx in range(self.encodings.shape[1]):
            cont_param = ContinuousParameter(
                name + "_" + str(column_idx),
                np.min(self.encodings[:, column_idx]),
                np.max(self.encodings[:, column_idx]),
            )
            self._cont_params.append(cont_param)

    def __str__(self):
        return f"<CategoricalParameter: {self.name} n_cat={self.dimension}>"

    def __repr__(self):
        return f"CategoricalParameter({self.name}, {self.encoding})"

    @property
    def encodings(self) -> np.ndarray:
        return self.encoding.encodings

    @property
    def model_parameters(self) -> List:
        return self._cont_params

    def round(self, x: np.ndarray) -> np.ndarray:
        return self.encoding.round(x)

    @property
    def bounds(self) -> List[Tuple]:
        """
        Returns a list of tuples containing where each tuple contains the minimum and maximum of the variables used to
        encode the categorical parameter..
        """
        return [(param.min, param.max) for param in self._cont_params]

    @property
    def dimension(self) -> int:
        return self.encodings.shape[1]

    def check_in_domain(self, x: np.ndarray) -> bool:
        """
        Verifies that given values lie within the parameter's domain

        :param x: 2d numpy array with shape (points, encoding) of points to check
        :return: A boolean value which indicates whether all points lie in the domain
        """
        if x.ndim != 2 or x.shape[1] != self.dimension:
            raise ValueError("Expected x shape (points, {}), actual is {}".format(self.dimension, x.shape))

        for i, param in enumerate(self._cont_params):
            # First check if this particular parameter is in domain
            param_in_domain = param.check_in_domain(x[:, i])
            if not param_in_domain:
                return False

        return True

    def sample_uniform(self, point_count: int) -> np.ndarray:
        """
        Generates multiple uniformly distributed random parameter points.

        :param point_count: number of data points to generate.
        :returns: Generated points with shape (point_count, num_features)
        """
        indices = np.random.randint(0, self.encodings.shape[0], point_count)
        return self.encodings[indices]
